from .addrevert import AddRevert
from .checkcellmetadata import CheckCellMetadata
from .clearalwayshiddentests import ClearAlwaysHiddenTests
from .clearhiddentests import ClearHiddenTests
from .clearmarkingscheme import ClearMarkScheme
from .clearoutput import ClearOutput
from .clearsolutions import ClearSolutions
from .computechecksums import ComputeChecksums
from .deduplicateids import DeduplicateIds
from .execute import Execute
from .getgrades import GetGrades
from .headerfooter import IncludeHeaderFooter
from .limitoutput import LimitOutput
from .lockcells import LockCells
from .overwritecells import OverwriteCells
from .overwritekernelspec import OverwriteKernelspec
from .saveautogrades import SaveAutoGrades
from .savecells import SaveCells

__all__ = [
    "IncludeHeaderFooter",
    "LockCells",
    "ClearSolutions",
    "SaveAutoGrades",
    "ComputeChecksums",
    "SaveCells",
    "OverwriteCells",
    "CheckCellMetadata",
    "Execute",
    "GetGrades",
    "ClearOutput",
    "LimitOutput",
    "DeduplicateIds",
    "ClearHiddenTests",
    "ClearMarkScheme",
    "OverwriteKernelspec",
    "AddRevert",
    "ClearAlwaysHiddenTests",
]



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from http import HTTPStatus

import tornado
from sqlalchemy.orm import joinedload
from sqlalchemy.orm.exc import ObjectDeletedError
from tornado.web import HTTPError

from grader_service.api.models.lecture import Lecture as LectureModel
from grader_service.handlers.base_handler import GraderBaseHandler, authorize
from grader_service.orm.assignment import Assignment
from grader_service.orm.base import DeleteState
from grader_service.orm.lecture import Lecture, LectureState
from grader_service.orm.takepart import Role, Scope
from grader_service.registry import VersionSpecifier, register_handler


@register_handler(r"\/api\/lectures\/?", VersionSpecifier.ALL)
class LectureBaseHandler(GraderBaseHandler):
    """
    Tornado Handler class for http requests to /lectures.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self):
        """
        Returns all lectures the user can access.
        """
        self.validate_parameters("complete")
        complete = self.get_argument("complete", None)
        if complete is None:  # return both complete and active lectures
            lectures = sorted(
                [
                    role.lecture
                    for role in self.user.roles
                    if role.lecture.deleted == DeleteState.active
                ],
                key=lambda lecture: lecture.id,
            )
        else:  # return either only complete or only active lectures
            state = LectureState.complete if (complete == "true") else LectureState.active
            lectures = sorted(
                [
                    role.lecture
                    for role in self.user.roles
                    if role.lecture.state == state and role.lecture.deleted == DeleteState.active
                ],
                key=lambda lecture: lecture.id,
            )

        self.write_json(lectures)

    @authorize([Scope.instructor])
    async def post(self):
        """
        Creates a new lecture from a "ghost"-lecture.

        :raises HTTPError: throws err if "ghost"-lecture was not found
        """
        self.validate_parameters()
        body = tornado.escape.json_decode(self.request.body)
        lecture_model = LectureModel.from_dict(body)

        lecture = (
            self.session.query(Lecture).filter(Lecture.code == lecture_model.code).one_or_none()
        )
        if lecture is None:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Lecture template not found")

        lecture.name = lecture_model.name
        lecture.code = lecture_model.code
        lecture.state = LectureState.complete if lecture_model.complete else LectureState.active
        lecture.deleted = DeleteState.active

        self.session.commit()
        self.set_status(HTTPStatus.CREATED)
        self.write_json(lecture)


@register_handler(r"\/api\/lectures\/(?P<lecture_id>\d*)\/?", VersionSpecifier.ALL)
class LectureObjectHandler(GraderBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}.
    """

    @authorize([Scope.instructor])
    async def put(self, lecture_id: int):
        """
        Updates a lecture.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        """
        self.validate_parameters()
        body = tornado.escape.json_decode(self.request.body)
        lecture_model = LectureModel.from_dict(body)
        lecture = self.session.get(Lecture, lecture_id)

        lecture.name = lecture_model.name
        lecture.state = LectureState.complete if lecture_model.complete else LectureState.active

        self.session.commit()
        self.write_json(lecture)

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int):
        """
        Finds lecture with the given lecture id.
        :param lecture_id: id of lecture
        :return: lecture with given id
        """
        self.validate_parameters()
        role = self.get_role(lecture_id)
        if role.lecture.deleted == DeleteState.deleted:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Lecture was not found")

        self.write_json(role.lecture)

    @authorize([Scope.instructor])
    async def delete(self, lecture_id: int):
        """
        "Soft"-delete a lecture.
        Softdeleting: lecture is still saved in the datastore
        but the users have not access to it.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :raises HTTPError: throws err if lecture was already deleted
        or was not found

        """
        self.validate_parameters()
        try:
            lecture = self.session.get(Lecture, lecture_id)
            if lecture.deleted == 1:
                raise HTTPError(404)
            lecture.deleted = 1
            a: Assignment
            for a in lecture.assignments:
                if (len(a.submissions)) > 0:
                    self.session.rollback()
                    raise HTTPError(
                        HTTPStatus.CONFLICT, "Cannot delete assignment because it has submissions"
                    )
                if a.status in ["released", "complete"]:
                    self.session.rollback()
                    raise HTTPError(
                        HTTPStatus.CONFLICT,
                        "Cannot delete assignment because its status is not created",
                    )

                a.deleted = 1
            self.session.commit()
        except ObjectDeletedError:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Lecture was not found")
        self.write("OK")


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/users\/?", version_specifier=VersionSpecifier.ALL
)
class LectureStudentsHandler(GraderBaseHandler):
    """
    Tornado Handler class for http requests to /lectures/{lecture_id}/users.
    """

    @authorize([Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int):
        """
        Finds all users of a lecture and groups them by roles.

        :param lecture_id: id of the lecture
        :return: a dictionary of user, tutor and instructor names lists
        """
        students = (
            self.session.query(Role)
            .options(joinedload(Role.user))
            .filter(Role.lectid == lecture_id)
            .filter(Role.role == Scope.student)
        )
        tutors = (
            self.session.query(Role)
            .options(joinedload(Role.user))
            .filter(Role.lectid == lecture_id)
            .filter(Role.role == Scope.tutor)
        )
        instructors = (
            self.session.query(Role)
            .options(joinedload(Role.user))
            .filter(Role.lectid == lecture_id)
            .filter(Role.role == Scope.instructor)
        )

        students = [
            {"id": s.user.id, "name": s.user.name, "display_name": s.user.display_name}
            for s in students
        ]
        tutors = [
            {"id": t.user.id, "name": t.user.name, "display_name": t.user.display_name}
            for t in tutors
        ]
        instructors = [
            {"id": i.user.id, "name": i.user.name, "display_name": i.user.display_name}
            for i in instructors
        ]

        counts = {"instructors": instructors, "tutors": tutors, "students": students}
        self.write_json(counts)



import shlex
import sys
from shutil import which
from subprocess import PIPE, STDOUT, Popen

from traitlets import Bool, Dict, List, Set, default

from grader_service.auth.auth import Authenticator
from grader_service.utils import maybe_future


class LocalAuthenticator(Authenticator):
    """Base class for Authenticators that work with local Linux/UNIX users

    Checks for local users, and can attempt to create them if they exist.
    """

    create_system_users = Bool(
        False,
        help="""
        If set to True, will attempt to create local system users if they do not exist already.

        Supports Linux and BSD variants only.
        """,
    ).tag(config=True)

    add_user_cmd = List(
        help="""
        The command to use for creating users as a list of strings

        For each element in the list, the string USERNAME will be replaced with
        the user's username. The username will also be appended as the final argument.

        For Linux, the default value is:

            ['adduser', '-q', '--gecos', '""', '--disabled-password']

        To specify a custom home directory, set this to:

            ['adduser', '-q', '--gecos', '""', '--home', '/customhome/USERNAME', '--disabled-password']

        This will run the command:

            adduser -q --gecos "" --home /customhome/river --disabled-password river

        when the user 'river' is created.
        """
    ).tag(config=True)

    @default("add_user_cmd")
    def _add_user_cmd_default(self):
        """Guess the most likely-to-work adduser command for each platform"""
        if sys.platform == "darwin":
            raise ValueError("I don't know how to create users on macOS")
        elif which("pw"):
            # Probably BSD
            return ["pw", "useradd", "-m", "-n"]
        else:
            # This appears to be the Linux non-interactive adduser command:
            return ["adduser", "-q", "--gecos", '""', "--disabled-password"]

    uids = Dict(
        help="""
        Dictionary of uids to use at user creation time.
        This helps ensure that users created from the database
        get the same uid each time they are created
        in temporary deployments or containers.
        """
    ).tag(config=True)

    allowed_groups = Set(
        help="""
        Allow login from all users in these UNIX groups.
        """
    ).tag(config=True, allow_config=True)

    def check_allowed(self, username, authentication=None):
        if self.check_allowed_groups(username, authentication):
            return True
        return super().check_allowed(username, authentication)

    def check_allowed_groups(self, username, authentication=None):
        """
        If allowed_groups is configured, check if authenticating user is part of group.
        """
        if not self.allowed_groups:
            return False
        for grnam in self.allowed_groups:
            try:
                group = self._getgrnam(grnam)
            except KeyError:
                self.log.error("No such group: [%s]" % grnam)
                continue
            if username in group.gr_mem:
                return True
        return False

    async def add_user(self, user):
        """Hook called whenever a new user is added

        If self.create_system_users, the user will attempt to be created if it doesn't exist.
        """
        user_exists = await maybe_future(self.system_user_exists(user))
        if not user_exists:
            if self.create_system_users:
                await maybe_future(self.add_system_user(user))
            else:
                raise KeyError(
                    f"User {user.name} does not exist on the system."
                    " Set LocalAuthenticator.create_system_users=True"
                    " to automatically create system users from jupyterhub users."
                )

        await maybe_future(super().add_user(user))

    @staticmethod
    def _getgrnam(name):
        """Wrapper function to protect against `grp` not being available
        on Windows
        """
        import grp

        return grp.getgrnam(name)

    @staticmethod
    def _getpwnam(name):
        """Wrapper function to protect against `pwd` not being available
        on Windows
        """
        import pwd

        return pwd.getpwnam(name)

    @staticmethod
    def _getgrouplist(name, group):
        """Wrapper function to protect against `os._getgrouplist` not being available
        on Windows
        """
        import os

        return os.getgrouplist(name, group)

    def system_user_exists(self, user):
        """Check if the user exists on the system"""
        try:
            self._getpwnam(user.name)
        except KeyError:
            return False
        else:
            return True

    def add_system_user(self, user):
        """Create a new local UNIX user on the system.

        Tested to work on FreeBSD and Linux, at least.
        """
        name = user.name
        cmd = [arg.replace("USERNAME", name) for arg in self.add_user_cmd]
        try:
            uid = self.uids[name]
            cmd += ["--uid", "%d" % uid]
        except KeyError:
            self.log.debug("No UID for user %s" % name)
        cmd += [name]
        self.log.info("Creating user: %s", " ".join(map(shlex.quote, cmd)))
        p = Popen(cmd, stdout=PIPE, stderr=STDOUT)
        p.wait()
        if p.returncode:
            err = p.stdout.read().decode("utf8", "replace")
            raise RuntimeError(f"Failed to create system user {name}: {err}")



import os

from grader_service.auth.auth import Authenticator
from grader_service.autograding.kube.kube_grader import KubeAutogradeExecutor
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.handlers.base_handler import BaseHandler
from grader_service.orm import User, Lecture
from grader_service.orm.base import DeleteState
from grader_service.orm.lecture import LectureState
from grader_service.orm.takepart import Scope, Role

# TODO: Only use DummyAuthenticator

# from grader_service.autograding.docker_grader import DockerAutogradeExecutor

print("### loading service config")

c.GraderService.service_host = "127.0.0.1"
# existing directory to use as the base directory for the grader service
service_dir = os.path.join(os.getcwd(), "service_dir")
c.GraderService.grader_service_dir = service_dir

c.GraderService.log_level = "DEBUG"

c.RequestHandlerConfig.autograde_executor_class = LocalAutogradeExecutor
c.LocalAutogradeExecutor.default_cell_timeout = 200

c.CeleryApp.conf = dict(
    broker_url="amqp://localhost",
    result_backend="rpc://",
    task_serializer="json",
    result_serializer="json",
    accept_content=["json"],
    broker_connection_retry_on_startup=True,
    task_always_eager=True,
)

c.CeleryApp.worker_kwargs = dict(concurrency=1, pool="prefork")


# JupyterHub client config
c.GraderService.oauth_clients = [
    {
        "client_id": "my_id",
        "client_secret": "my_secret",
        "redirect_uri": "http://localhost:8080/hub/oauth_callback",
    }
]

from grader_service.auth.dummy import DummyAuthenticator

c.GraderService.authenticator_class = DummyAuthenticator
c.Authenticator.allowed_users = {"admin", "instructor", "student", "tutor"}
c.Authenticator.admin_users = {"admin"}
c.GraderService.load_roles = {
    "lecture1": [
        {"members": ["student"], "role": "student"},
        {"members": ["tutor"], "role": "tutor"},
        {"members": ["instructor", "admin"], "role": "instructor"},
    ]
}



import base64
import json
import os
from binascii import a2b_hex
from concurrent.futures import ThreadPoolExecutor

from traitlets import Any, Integer, List, default, observe, validate
from traitlets.config import Config, SingletonConfigurable

try:
    import cryptography
    from cryptography.fernet import Fernet, InvalidToken, MultiFernet
except ImportError:
    cryptography = None

    class InvalidToken(Exception):
        pass


from grader_service.utils import maybe_future

KEY_ENV = "JUPYTERHUB_CRYPT_KEY"


class EncryptionUnavailable(Exception):
    pass


class CryptographyUnavailable(EncryptionUnavailable):
    def __str__(self):
        return "cryptography library is required for encryption"


class NoEncryptionKeys(EncryptionUnavailable):
    def __str__(self):
        return "Encryption keys must be specified in %s env" % KEY_ENV


def _validate_key(key):
    """Validate and return a 32B key

    Args:
    key (bytes): The key to be validated.
        Can be:
        - base64-encoded (44 bytes)
        - hex-encoded (64 bytes)
        - raw 32 byte key

    Returns:
    key (bytes): raw 32B key
    """
    if isinstance(key, str):
        key = key.encode("ascii")

    if len(key) == 44:
        try:
            key = base64.urlsafe_b64decode(key)
        except ValueError:
            pass

    elif len(key) == 64:
        try:
            # 64B could be 32B, hex-encoded
            return a2b_hex(key)
        except ValueError:
            # not 32B hex
            pass

    if len(key) != 32:
        raise ValueError("Encryption keys must be 32 bytes, hex or base64-encoded.")

    return key


class CryptKeeper(SingletonConfigurable):
    """Encapsulate encryption configuration

    Use via the encryption_config singleton below.
    """

    n_threads = Integer(
        max(os.cpu_count(), 1), config=True, help="The number of threads to allocate for encryption"
    )

    @default("config")
    def _config_default(self):
        # load application config by default
        from .app import JupyterHub

        if JupyterHub.initialized():
            return JupyterHub.instance().config
        else:
            return Config()

    executor = Any()

    def _executor_default(self):
        return ThreadPoolExecutor(self.n_threads)

    keys = List(config=True)

    def _keys_default(self):
        if KEY_ENV not in os.environ:
            return []
        # key can be a ;-separated sequence for key rotation.
        # First item in the list is used for encryption.
        return [_validate_key(key) for key in os.environ[KEY_ENV].split(";") if key.strip()]

    @validate("keys")
    def _ensure_bytes(self, proposal):
        # cast str to bytes
        return [_validate_key(key) for key in proposal.value]

    fernet = Any()

    def _fernet_default(self):
        if cryptography is None or not self.keys:
            return None
        return MultiFernet([Fernet(base64.urlsafe_b64encode(key)) for key in self.keys])

    @observe("keys")
    def _update_fernet(self, change):
        self.fernet = self._fernet_default()

    def check_available(self):
        if cryptography is None:
            raise CryptographyUnavailable()
        if not self.keys:
            raise NoEncryptionKeys()

    def _encrypt(self, data):
        """Actually do the encryption. Runs in a background thread.

        data is serialized to bytes with pickle.
        bytes are returned.
        """
        return self.fernet.encrypt(json.dumps(data).encode("utf8"))

    def encrypt(self, data):
        """Encrypt an object with cryptography"""
        self.check_available()
        return maybe_future(self.executor.submit(self._encrypt, data))

    def _decrypt(self, encrypted):
        decrypted = self.fernet.decrypt(encrypted)
        return json.loads(decrypted.decode("utf8"))

    def decrypt(self, encrypted):
        """Decrypt an object with cryptography"""
        self.check_available()
        return maybe_future(self.executor.submit(self._decrypt, encrypted))


def encrypt(data):
    """encrypt some data with the crypt keeper.

    data will be serialized with pickle.
    Returns a Future whose result will be bytes.
    """
    return CryptKeeper.instance().encrypt(data)


def decrypt(data):
    """decrypt some data with the crypt keeper

    Returns a Future whose result will be the decrypted, deserialized data.
    """
    return CryptKeeper.instance().decrypt(data)



import asyncio
from typing import Union

from celery import Celery, Task
from tornado.web import HTTPError

from grader_service.autograding.celery.app import CeleryApp
from grader_service.autograding.local_feedback import LocalFeedbackExecutor
from grader_service.handlers.base_handler import RequestHandlerConfig
from grader_service.orm.submission import FeedbackStatus, Submission
from grader_service.plugins.lti import LTISyncGrades

# Note: The celery instance is lazy so we can still add configuration later
app = Celery(set_as_current=True)


class GraderTask(Task):
    def __init__(self) -> None:
        self.celery = CeleryApp.instance()
        self.log = self.celery.log
        self.Session = self.celery.sessionmaker
        self._sessions = {}

    def before_start(self, task_id, args, kwargs):
        self._sessions[task_id] = self.Session()
        super().before_start(task_id, args, kwargs)

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        session = self._sessions.pop(task_id)
        session.close()
        super().after_return(status, retval, task_id, args, kwargs, einfo)

    @property
    def session(self):
        return self._sessions[self.request.id]


@app.task(bind=True, base=GraderTask)
def add(self: GraderTask, x, y):
    print(type(self.session))
    print("Adding {} and {}".format(x, y))
    return x + y


@app.task(bind=True, base=GraderTask)
def autograde_task(self: GraderTask, lecture_id: int, assignment_id: int, sub_id: int):
    from grader_service.main import GraderService

    grader_service_dir = GraderService.instance().grader_service_dir

    submission = self.session.get(Submission, sub_id)
    if submission is None:
        raise ValueError("Submission not found")
    if submission.assignment.id != assignment_id or submission.assignment.lecture.id != lecture_id:
        raise ValueError(
            f"invalid submission {submission.id}: {assignment_id=:}, {lecture_id=:} does not match"
        )

    executor = RequestHandlerConfig.instance().autograde_executor_class(
        grader_service_dir, submission, config=self.celery.config
    )
    self.log.info(f"Running autograding task for submission {submission.id}")
    executor.start()
    self.log.info(f"Autograding task of submission {submission.id} exited!")


@app.task(bind=True, base=GraderTask)
def generate_feedback_task(self: GraderTask, lecture_id: int, assignment_id: int, sub_id: int):
    from grader_service.main import GraderService

    grader_service_dir = GraderService(config=self.celery.config).grader_service_dir

    submission = self.session.get(Submission, sub_id)
    if submission is None:
        raise ValueError("Submission not found")
    if submission.assignment.id != assignment_id or submission.assignment.lecture.id != lecture_id:
        raise ValueError(
            f"invalid submission {submission.id}: {assignment_id=:}, {lecture_id=:} does not match"
        )

    executor = LocalFeedbackExecutor(grader_service_dir, submission, config=self.celery.config)
    executor.start()
    if submission.feedback_status == FeedbackStatus.GENERATED:
        self.log.info("Successfully generated feedback for submission %s!", submission.id)
    else:
        self.log.error("Failed to generate feedback for submission %s!", submission.id)


@app.task(bind=True, base=GraderTask)
def lti_sync_task(
    self: GraderTask,
    lecture: dict,
    assignment: dict,
    submissions: list[dict],
    feedback_sync: bool = False,
) -> Union[dict, None]:
    """Gathers submissions based on params and starts LTI sync process
    :param lecture: lecture object
    :param assignment: assignment object
    :param submissions: submissions to be synced (including user information)
    :param feedback_sync(optional): if True, the sync task was started by a feedback generation
    """
    lti_plugin = LTISyncGrades.instance()
    # check if the lti plugin is enabled
    if lti_plugin.check_if_lti_enabled(
        lecture, assignment, submissions, feedback_sync=feedback_sync
    ):
        try:
            results = asyncio.run(lti_plugin.start(lecture, assignment, submissions))
            return results
        except HTTPError as e:
            err_msg = f"Could not sync grades: {e.reason}"
            self.log.info(err_msg)
            raise e
        except Exception as e:
            self.log.error("Could not sync grades: " + str(e))
            raise HTTPError(500, reason="An unexpected error occured.")
    else:
        # if the synchronisation task is automatic only log event
        if feedback_sync:
            self.log.info("Skipping LTI grade synchronisation, because it is not enabled")
        else:
            # else tell the user that the plugin is disabled
            raise HTTPError(403, reason="LTI plugin is not enabled by administator.")
    return None



import os
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.escape import json_encode, json_decode

## generic
c.JupyterHub.admin_access = True
c.Spawner.default_url = "/lab"
c.Spawner.cmd = ["jupyter-labhub"]

## authenticator
from jupyterhub.auth import DummyAuthenticator

c.JupyterHub.authenticator_class = DummyAuthenticator
c.Authenticator.allowed_users = {"admin", "instructor", "tutor", "student"}
c.Authenticator.admin_users = {"admin"}
c.JupyterHub.load_groups = {
    "lect1:instructor": {"users": ["admin", "instructor"]},
    "lect1:student": {"users": ["student"]},
    "lect1:tutor": {"users": ["tutor"]},
}
c.Authenticator.enable_auth_state = True

## spawner
c.JupyterHub.spawner_class = "jupyterhub.spawner.SimpleLocalProcessSpawner"
c.SimpleLocalProcessSpawner.home_dir_template = os.path.join(os.getcwd(), "home_dir", "{username}")


async def pre_spawn_hook(spawner):
    http_client = AsyncHTTPClient()
    data = {"token": spawner.api_token}
    request = HTTPRequest(
        url="http://localhost:4010/services/grader/login", method="POST", body=json_encode(data)
    )

    response = await http_client.fetch(request=request)
    grader_api_token = json_decode(response.body)["api_token"]
    spawner.environment.update({"GRADER_API_TOKEN": grader_api_token})


c.Spawner.pre_spawn_hook = pre_spawn_hook

## simple setup
c.JupyterHub.ip = "127.0.0.1"
c.JupyterHub.port = 8080

c.JupyterHub.services.append(
    {
        "name": "grader",
        "url": "http://127.0.0.1:4010",
        "api_token": "7572f93a2e7640999427d9289c8318c0",
    }
)

c.JupyterHub.log_level = "INFO"



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import json
import os
from typing import Callable, Dict, Optional, Union
from urllib.parse import ParseResultBytes, quote_plus, urlencode, urlparse

from tornado.escape import json_decode
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPResponse
from traitlets.config.configurable import SingletonConfigurable
from traitlets.traitlets import TraitError, Unicode, validate


class RequestService(SingletonConfigurable):
    url = Unicode(os.environ.get("GRADER_HOST_URL", "http://127.0.0.1:4010"))

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.http_client = AsyncHTTPClient()
        self._service_cookie = None

    async def request(
        self,
        method: str,
        endpoint: str,
        body: Union[dict, str] = None,
        header: Dict[str, str] = None,
        decode_response: bool = True,
        request_timeout: float = 20.0,
        connect_timeout: float = 20.0,
        response_callback: Optional[Callable[[HTTPResponse], None]] = None,
    ) -> Union[dict, list, HTTPResponse]:
        self.log.info(self.url + endpoint)
        if self._service_cookie:
            header["Cookie"] = self._service_cookie

        if isinstance(body, dict):
            body = json.dumps(body)

        # Build HTTPRequest
        request = HTTPRequest(
            url=self.url + endpoint,
            method=method,
            headers=header,
            request_timeout=request_timeout,
            connect_timeout=connect_timeout,
        )
        # Add body if exists
        if body:
            request.body = body

        # Sent HTTPRequest
        response: HTTPResponse = await self.http_client.fetch(request=request)

        for cookie in response.headers.get_list("Set-Cookie"):
            token = header.get("Authorization", None)
            if token and token.startswith("Token "):
                token = token[len("Token ") :]
            else:
                continue
            if cookie.startswith(token):
                self._service_cookie = cookie

        if response_callback:
            response_callback(response)

        if decode_response:
            return json_decode(response.body)
        else:
            return response

    @validate("url")
    def _validate_url(self, proposal):
        url = proposal["value"]
        result: ParseResultBytes = urlparse(url)
        if not all([result.scheme, result.hostname]):
            raise TraitError("Invalid url: at least has to contain scheme and hostname")
        return url

    @staticmethod
    def get_query_string(params: dict) -> str:
        d = {k: v for k, v in params.items() if v is not None}
        query_params: str = urlencode(d, quote_via=quote_plus)
        return "?" + query_params if query_params != "" else ""



from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from traitlets import List

from grader_service.convert import utils
from grader_service.convert.gradebook.gradebook import Gradebook, MissingEntry
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class GetGrades(NbGraderPreprocessor):
    """Preprocessor for saving grades from the database to the notebook"""

    display_data_priority = List(
        [
            "text/html",
            "application/pdf",
            "text/latex",
            "image/svg+xml",
            "image/png",
            "image/jpeg",
            "text/plain",
        ],
        config=True,
    )

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # pull information from the resources
        self.notebook_id = resources["unique_key"]
        self.json_path = resources["output_json_path"]

        # connect to the database
        self.gradebook = Gradebook(self.json_path)

        with self.gradebook:
            # process the cells
            nb, resources = super(GetGrades, self).preprocess(nb, resources)

            try:
                notebook = self.gradebook.find_notebook(self.notebook_id)
            except MissingEntry:
                self.log.info(
                    "Did not find a gradebook entry for the notebook %s", self.notebook_id
                )
                raise

            resources["nbgrader"]["score"] = notebook.score
            resources["nbgrader"]["max_score"] = notebook.max_score

        return nb, resources

    def _get_comment(self, cell: NotebookNode, resources: ResourcesDict) -> None:
        """Graders can optionally add comments to the student's solutions, so
        add the comment information into the database if it doesn't
        already exist. It should NOT overwrite existing comments that
        might have been added by a grader already.

        """

        # retrieve or create the comment object from the database
        comment = self.gradebook.find_comment(
            cell.metadata["nbgrader"]["grade_id"], self.notebook_id
        )

        # save it in the notebook
        cell.metadata.nbgrader["comment"] = comment.comment

    def _get_score(self, cell: NotebookNode, resources: ResourcesDict) -> None:
        grade = self.gradebook.find_grade(cell.metadata["nbgrader"]["grade_id"], self.notebook_id)

        cell.metadata.nbgrader["score"] = grade.score
        cell.metadata.nbgrader["points"] = grade.max_score

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # if it's a solution cell, then add a comment
        if utils.is_solution(cell):
            self._get_comment(cell, resources)

        # if it's a grade cell, the add a grade
        if utils.is_grade(cell):
            self._get_score(cell, resources)
        # if it's a task cell, then add a comment and a score
        if utils.is_task(cell):
            self._get_comment(cell, resources)
            self._get_score(cell, resources)

        return cell, resources



from sqlalchemy import Column, Integer, Unicode
from sqlalchemy.orm import relationship

from grader_service.orm.api_token import APIToken
from grader_service.orm.base import Base
from grader_service.orm.oauthcode import OAuthCode


class OAuthClient(Base):
    __tablename__ = "oauth_client"
    id = Column(Integer, primary_key=True, autoincrement=True)
    identifier = Column(Unicode(255), unique=True)
    description = Column(Unicode(1023))
    secret = Column(Unicode(255))
    redirect_uri = Column(Unicode(1023))

    @property
    def client_id(self):
        return self.identifier

    access_tokens = relationship(
        APIToken, back_populates="oauth_client", cascade="all, delete-orphan", passive_deletes=True
    )
    codes = relationship(
        OAuthCode, back_populates="client", cascade="all, delete-orphan", passive_deletes=True
    )

    def __repr__(self):
        return f"<{self.__class__.__name__}(identifier={self.identifier!r})>"



import traceback
from typing import Dict, Tuple

from nbformat.notebooknode import NotebookNode

from grader_service.convert.nbgraderformat import MetadataValidator, ValidationError
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class CheckCellMetadata(NbGraderPreprocessor):
    """A preprocessor for checking that grade ids are unique."""

    def preprocess(self, nb: NotebookNode, resources: Dict) -> Tuple[NotebookNode, Dict]:
        try:
            MetadataValidator().validate_nb(nb)
        except ValidationError as e:
            self.log.error(traceback.format_exc())
            msg = "Notebook failed to validate: " + e.message
            self.log.error(msg)
            raise ValidationError(msg)

        return nb, resources



import datetime
import json
import os
import shutil
from http import HTTPStatus
from pathlib import Path
from typing import Union

import isodate
import tornado
from sqlalchemy.exc import IntegrityError
from tornado.web import HTTPError

from grader_service.api.models.assignment import Assignment as AssignmentModel
from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.convert.gradebook.models import GradeBookModel
from grader_service.handlers.base_handler import GraderBaseHandler, authorize
from grader_service.handlers.submissions import SubmissionHandler
from grader_service.orm.assignment import Assignment
from grader_service.orm.base import DeleteState
from grader_service.orm.submission import Submission
from grader_service.orm.takepart import Role, Scope
from grader_service.registry import VersionSpecifier, register_handler

from .handler_utils import GitRepoType, parse_ids


def validate_assignment_settings(settings: Union[AssignmentSettings, None]):
    if settings is None:
        return
    late_submission = settings.late_submission or []
    current_period = isodate.parse_duration("P0D")
    curr_scaling = 1.0
    for period in late_submission:
        try:
            d = isodate.parse_duration(period.period)
            if d < datetime.timedelta(0):
                raise ValueError
            if not isinstance(period.scaling, float):
                raise TypeError
            s = period.scaling
        except (isodate.isoerror.ISO8601Error, TypeError, ValueError):
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Invalid settings!")
        if d <= current_period:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Period lengths are not increasing!)")
        if s <= 0.0 or s >= 1.0:
            raise HTTPError(
                HTTPStatus.BAD_REQUEST,
                reason="Score scaling has to be between 0.0 and 1.0 exclusive!",
            )
        if s >= curr_scaling:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Score scaling is not decreasing!")
        current_period = d
        curr_scaling = s


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/?",
    version_specifier=VersionSpecifier.ALL,
)
class AssignmentBaseHandler(GraderBaseHandler):
    """Tornado Handler for http requests

    route: /lectures/{lecture_id}/assignments."""

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int):
        """Returns all assignments of a lecture.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :raises HTTPError: throws err if lecture is deleted
        """
        lecture_id = parse_ids(lecture_id)
        self.validate_parameters("include-submissions")
        role = self.get_role(lecture_id)
        if role.lecture.deleted == DeleteState.deleted:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Lecture not found")

        if role.role == Scope.student:  # students do not get assignments that are created
            assignments = (
                self.session.query(Assignment)
                .filter(
                    Assignment.lectid == role.lecture.id,
                    Assignment.deleted == DeleteState.active,
                    Assignment.status != "created",
                    Assignment.status != "pushed",
                )
                .all()
            )

        else:
            assignments = [a for a in role.lecture.assignments if (a.deleted == DeleteState.active)]

        # Handle the case that the user wants to include submissions
        include_submissions = self.get_argument("include-submissions", "true") == "true"
        if include_submissions:
            assignids = [a.id for a in assignments]
            user_id = self.user.id
            results = (
                self.session.query(Submission)
                .filter(Submission.assignid.in_(assignids), Submission.user_id == user_id)
                .all()
            )
            # Create a combined list of assignments and submissions
            assignments = [a.serialize() for a in assignments]
            submissions = [s.serialize() for s in results]
            for a in assignments:
                a["submissions"] = [s for s in submissions if s["assignid"] == a["id"]]
        assignments = sorted(assignments, key=lambda o: o["id"])
        self.write_json(assignments)

    @authorize([Scope.instructor])
    async def post(self, lecture_id: int):
        """Creates a new assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        """
        lecture_id = parse_ids(lecture_id)
        self.validate_parameters()
        role = self.get_role(lecture_id)
        if role.lecture.deleted == DeleteState.deleted:
            raise HTTPError(HTTPStatus.NOT_FOUND)
        try:
            body = tornado.escape.json_decode(self.request.body)
            assignment_model = AssignmentModel.from_dict(body)
        except ValueError as e:
            # TODO Return useful error message
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason=str(e))
        assignment = Assignment()

        assignment.name = assignment_model.name
        assignment_with_name = (
            self.session.query(Assignment)
            .filter(
                Assignment.name == assignment.name,
                Assignment.deleted == DeleteState.active,
                Assignment.lectid == lecture_id,
            )
            .one_or_none()
        )

        if assignment_with_name is not None:
            raise HTTPError(HTTPStatus.CONFLICT, reason="Assignment name is already being used")
        if assignment_model.settings is None:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Assignment missing settings key")
        if (assignment_model.settings.max_submissions is not None) and (
            assignment_model.settings.max_submissions < 1
        ):
            msg = "Maximum number of submissions cannot be smaller than 1!"
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason=msg)
        validate_assignment_settings(assignment_model.settings)

        assignment.lectid = lecture_id
        assignment.status = assignment_model.status
        assignment.points = 0
        assignment.deleted = DeleteState.active
        assignment.settings = assignment_model.settings

        self.session.add(assignment)
        try:
            self.session.commit()
        except IntegrityError as e:
            self.log.error(e)
            self.session.rollback()
            raise HTTPError(
                HTTPStatus.UNPROCESSABLE_ENTITY, reason="Cannot add object to database."
            )
        self.set_status(HTTPStatus.CREATED)
        self.write_json(assignment)


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/?",
    version_specifier=VersionSpecifier.ALL,
)
class AssignmentObjectHandler(GraderBaseHandler):
    """Tornado Handler for http requests

    route: /lectures/{lecture_id}/assignments/{assignment_id}."""

    @authorize([Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int):
        """Updates an assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters("recalc-scores")
        recalc_scores = self.get_argument("recalc-scores", None) == "true"
        body = tornado.escape.json_decode(self.request.body)
        assignment_model = AssignmentModel.from_dict(body)
        assignment = self.get_assignment(lecture_id, assignment_id)
        # Validate name
        assignment_with_name = (
            self.session.query(Assignment)
            .filter(
                Assignment.name == assignment_model.name,
                Assignment.deleted == DeleteState.active,
                Assignment.lectid == assignment.lectid,
            )
            .one_or_none()
        )

        if (assignment_with_name is not None) and (assignment_with_name.id != assignment_id):
            raise HTTPError(HTTPStatus.CONFLICT, reason="Assignment name is already being used")
        if (assignment_model.settings.max_submissions is not None) and (
            assignment_model.settings.max_submissions < 1
        ):
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Max num submissions < 1!")
        validate_assignment_settings(assignment_model.settings)

        assignment.name = assignment_model.name
        assignment.status = assignment_model.status
        if (assignment_model.settings.autograde_type == "full_auto") and (
            assignment.properties is not None
        ):
            model = GradeBookModel.from_dict(json.loads(assignment.properties))
            _check_full_auto_grading(self, model)

        assignment.settings = assignment_model.settings

        # recalculate scores of assignment submissions
        if recalc_scores:
            submissions = self.get_all_submissions(assignment_id)
            role = Role()
            role.role = Scope.instructor
            for sub in submissions:
                late_submission_scaling = SubmissionHandler.calculate_late_submission_scaling(
                    assignment, sub.date, role
                )
                # change score if the scaling changed
                if sub.score_scaling != late_submission_scaling:
                    sub.score = late_submission_scaling * sub.grading_score
                    sub.score_scaling = late_submission_scaling

        self.session.commit()
        self.write_json(assignment)

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """Returns a specific assignment of a lecture.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        assignment = self.get_assignment(lecture_id=lecture_id, assignment_id=assignment_id)
        self.write_json(assignment)

    @authorize([Scope.instructor])
    async def delete(self, lecture_id: int, assignment_id: int):
        """Soft-Deletes a specific assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if assignment was not found or deleted
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()
        assignment = self.get_assignment(lecture_id, assignment_id)

        if assignment.status in ["released", "complete"]:
            msg = "Cannot delete released or completed assignments!"
            raise HTTPError(HTTPStatus.CONFLICT, reason=msg)

        previously_deleted = (
            self.session.query(Assignment)
            .filter(
                Assignment.lectid == lecture_id,
                Assignment.name == assignment.name,
                Assignment.deleted == DeleteState.deleted,
            )
            .one_or_none()
        )
        if previously_deleted is not None:
            self.session.delete(previously_deleted)
            self.session.commit()

        assignment.deleted = DeleteState.deleted
        self.session.commit()


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/reset\/?",
    version_specifier=VersionSpecifier.ALL,
)
class AssignmentResetHandler(GraderBaseHandler):
    """Tornado Handler for http requests

    route: /lectures/{lecture_id}/assignments/{assignment_id}/reset."""

    @authorize([Scope.instructor, Scope.tutor, Scope.student])
    async def get(self, lecture_id: int, assignment_id: int):
        self.validate_parameters()
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        assignment = self.get_assignment(lecture_id, assignment_id)

        grader_service_dir = Path(self.application.grader_service_dir)
        git_path_base = grader_service_dir.joinpath(
            "tmp", assignment.lecture.code, assignment.name, self.user.name
        )
        self.log.info(git_path_base)
        # Deleting dir
        if os.path.exists(git_path_base):
            shutil.rmtree(git_path_base)

        self.log.info(f"DIR {git_path_base}")
        os.makedirs(git_path_base, exist_ok=True)
        git_path_release = os.path.join(git_path_base, "release")
        git_path_user = os.path.join(git_path_base, self.user.name)
        self.log.info(f"GIT BASE {git_path_base}")
        self.log.info(f"GIT RELEASE {git_path_release}")
        self.log.info(f"GIT USER {git_path_user}")

        repo_path_release = self.construct_git_dir(
            GitRepoType.RELEASE, assignment.lecture, assignment
        )
        repo_path_user = self.construct_git_dir(GitRepoType.USER, assignment.lecture, assignment)

        self.duplicate_release_repo(
            repo_path_release=repo_path_release,
            repo_path_user=repo_path_user,
            assignment=assignment,
            message="Reset Assignment",
        )

        self.write_json(assignment)


@register_handler(
    path=r"\/api\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)"
    r"\/properties\/?",
    version_specifier=VersionSpecifier.ALL,
)
class AssignmentPropertiesHandler(GraderBaseHandler):
    """Tornado Handler for http requests

    route: /lectures/{lecture_id}/assignments/{assignment_id}/properties."""

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """
        Returns the properties of a specific assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: if assignment or their properties not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)

        self.validate_parameters()
        assignment = self.get_assignment(lecture_id, assignment_id)
        if assignment.properties is not None:
            if self.get_role(lecture_id).role == Scope.student:
                model = GradeBookModel.from_dict(json.loads(assignment.properties))
                for notebook in model.notebooks.values():
                    notebook.source_cells_dict = {}
                self.write(json.dumps(model.to_dict()))
            else:
                self.write(assignment.properties)
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Assignment not found")

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int):
        """
        Updates the properties of a specific assignment.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if the assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()
        assignment = self.get_assignment(lecture_id, assignment_id)
        properties_string: str = self.request.body.decode("utf-8")

        model = GradeBookModel.from_dict(json.loads(properties_string))
        # Check if assignment contains no cells that
        # need manual grading if assignment is fully auto grade
        if assignment.settings.autograde_type == "full_auto":
            _check_full_auto_grading(self, model)

        assignment.properties = properties_string
        assignment.points = model.max_score
        self.session.commit()


def _check_full_auto_grading(self: GraderBaseHandler, model):
    """Check if the assignment notebook contains manually graded cells

    If cells are found, throw error.
    :param self: handler class
    :param model: the notebook which is being tested
    :return: void"""
    for nb in model.notebooks.values():
        if len(nb.task_cells_dict) > 0:
            msg = "Fully autograded notebook cannot contain task cells!"
            raise HTTPError(HTTPStatus.CONFLICT, reason=msg)
        grades = set(nb.grade_cells_dict.keys())
        solutions = set(nb.solution_cells_dict.keys())
        if len(grades & solutions) > 0:
            msg = "Fully autograded notebook cannot contain manually graded cells!"
            raise HTTPError(HTTPStatus.CONFLICT, reason=msg)



from .autograde import Autograde
from .base import BaseConverter
from .generate_assignment import GenerateAssignment
from .generate_feedback import GenerateFeedback

__all__ = ["BaseConverter", "GenerateAssignment", "Autograde", "GenerateFeedback"]



from concurrent.futures import ThreadPoolExecutor

from tornado.concurrent import run_on_executor
from traitlets import Any, Bool, Set, Unicode, default

from grader_service.auth.local import LocalAuthenticator

try:
    import pamela
except Exception as e:
    pamela = None
    _pamela_error = e


class PAMAuthenticator(LocalAuthenticator):
    """Authenticate local UNIX users with PAM"""

    # run PAM in a thread, since it can be slow
    executor = Any()

    @default("executor")
    def _default_executor(self):
        return ThreadPoolExecutor(1)

    encoding = Unicode(
        "utf8",
        help="""
        The text encoding to use when communicating with PAM
        """,
    ).tag(config=True)

    service = Unicode(
        "login",
        help="""
        The name of the PAM service to use for authentication
        """,
    ).tag(config=True)

    open_sessions = Bool(
        False,
        help="""
        Whether to open a new PAM session when spawners are started.

        This may trigger things like mounting shared filesystems,
        loading credentials, etc. depending on system configuration.

        The lifecycle of PAM sessions is not correct,
        so many PAM session configurations will not work.

        If any errors are encountered when opening/closing PAM sessions,
        this is automatically set to False.

        .. versionchanged:: 2.2

            Due to longstanding problems in the session lifecycle,
            this is now disabled by default.
            You may opt-in to opening sessions by setting this to True.
        """,
    ).tag(config=True)

    check_account = Bool(
        True,
        help="""
        Whether to check the user's account status via PAM during authentication.

        The PAM account stack performs non-authentication based account
        management. It is typically used to restrict/permit access to a
        service and this step is needed to access the host's user access control.

        Disabling this can be dangerous as authenticated but unauthorized users may
        be granted access and, therefore, arbitrary execution on the system.
        """,
    ).tag(config=True)

    admin_groups = Set(
        help="""
        Authoritative list of user groups that determine admin access.
        Users not in these groups can still be granted admin status through admin_users.

        allowed/blocked rules still apply.

        Note: As of JupyterHub 2.0,
        full admin rights should not be required,
        and more precise permissions can be managed via roles.
        """
    ).tag(config=True)

    pam_normalize_username = Bool(
        False,
        help="""
        Round-trip the username via PAM lookups to make sure it is unique

        PAM can accept multiple usernames that map to the same user,
        for example DOMAIN\\username in some cases.  To prevent this,
        convert username into uid, then back to uid to normalize.
        """,
    ).tag(config=True)

    def __init__(self, **kwargs):
        if pamela is None:
            raise _pamela_error from None
        super().__init__(**kwargs)

    @run_on_executor
    def is_admin(self, handler, authentication):
        """PAM admin status checker. Returns Bool to indicate user admin status."""
        # Checks upper level function (admin_users)
        admin_status = super().is_admin(handler, authentication)
        username = authentication["name"]

        # If not yet listed as an admin, and admin_groups is on, use it authoritatively
        if not admin_status and self.admin_groups:
            try:
                # Most likely source of error here is a group name <-> gid mapping failure
                # This is most likely due to a typo in the configuration or in the case of LDAP/AD, a network
                # connectivity issue. Maybe a long one where the local caches have timed out, though PAM would
                # most likely would refuse to authenticate a remote user by that point.

                # It was decided that the best course of action on group resolution failure was to
                # fail to authenticate and raise instead of soft-failing and not changing admin status
                # (returning None instead of just the username) as this indicates some sort of system failure

                admin_group_gids = {self._getgrnam(x).gr_gid for x in self.admin_groups}
                user_group_gids = set(self._getgrouplist(username, self._getpwnam(username).pw_gid))
                admin_status = len(admin_group_gids & user_group_gids) != 0

            except Exception as e:
                if handler is not None:
                    self.log.error(
                        "PAM Admin Group Check failed (%s@%s): %s",
                        username,
                        handler.request.remote_ip,
                        e,
                    )
                else:
                    self.log.error("PAM Admin Group Check failed: %s", e)
                # re-raise to return a 500 to the user and indicate a problem. We failed, not them.
                raise

        return admin_status

    @run_on_executor
    def authenticate(self, handler, data):
        """Authenticate with PAM, and return the username if login is successful.

        Return None otherwise.
        """
        username = data["username"]
        password = data["password"]
        if "otp" in data:
            # OTP given, pass as tuple (requires pamela 1.1)
            password = (data["password"], data["otp"])
        try:
            pamela.authenticate(username, password, service=self.service, encoding=self.encoding)
        except pamela.PAMError as e:
            if handler is not None:
                self.log.warning(
                    "PAM Authentication failed (%s@%s): %s", username, handler.request.remote_ip, e
                )
            else:
                self.log.warning("PAM Authentication failed: %s", e)
            return None

        if self.check_account:
            try:
                pamela.check_account(username, service=self.service, encoding=self.encoding)
            except pamela.PAMError as e:
                if handler is not None:
                    self.log.warning(
                        "PAM Account Check failed (%s@%s): %s",
                        username,
                        handler.request.remote_ip,
                        e,
                    )
                else:
                    self.log.warning("PAM Account Check failed: %s", e)
                return None

        return username

    @run_on_executor
    def pre_spawn_start(self, user, spawner):
        """Open PAM session for user if so configured"""
        if not self.open_sessions:
            return
        try:
            pamela.open_session(user.name, service=self.service, encoding=self.encoding)
        except pamela.PAMError as e:
            self.log.warning("Failed to open PAM session for %s: %s", user.name, e)
            self.log.warning("Disabling PAM sessions from now on.")
            self.open_sessions = False

    @run_on_executor
    def post_spawn_stop(self, user, spawner):
        """Close PAM session for user if we were configured to opened one"""
        if not self.open_sessions:
            return
        try:
            pamela.close_session(user.name, service=self.service, encoding=self.encoding)
        except pamela.PAMError as e:
            self.log.warning("Failed to close PAM session for %s: %s", user.name, e)
            self.log.warning("Disabling PAM sessions from now on.")
            self.open_sessions = False

    def normalize_username(self, username):
        """Round-trip the username to normalize it with PAM

        PAM can accept multiple usernames as the same user, normalize them."""
        if self.pam_normalize_username:
            import pwd

            uid = pwd.getpwnam(username).pw_uid
            username = pwd.getpwuid(uid).pw_name
            username = self.username_map.get(username, username)
            return username
        else:
            return super().normalize_username(username)



from sqlalchemy import Column, ForeignKey, Integer, Text
from sqlalchemy.orm import relationship

from grader_service.orm.base import Base, Serializable


class SubmissionLogs(Base, Serializable):
    __tablename__ = "submission_logs"
    sub_id = Column(Integer, ForeignKey("submission.id", ondelete="CASCADE"), primary_key=True)
    logs = Column(Text, nullable=True)

    submission = relationship("Submission", back_populates="logs")



"""Base API handlers"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import json
import warnings
from functools import lru_cache
from http.client import responses
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from sqlalchemy.exc import SQLAlchemyError
from tornado import web

from ..handlers.base_handler import BaseHandler
from ..utils import isoformat

PAGINATION_MEDIA_TYPE = "application/jupyterhub-pagination+json"


class APIHandler(BaseHandler):
    """Base class for API endpoints

    Differences from page handlers:

    - JSON responses and errors
    - strict content-security-policy
    - methods for REST API models
    """

    # accept token-based authentication for API requests
    _accept_token_auth = True

    @property
    def content_security_policy(self):
        return "; ".join([super().content_security_policy, "default-src 'none'"])

    def get_content_type(self):
        return "application/json"

    @property
    @lru_cache()
    def accepts_pagination(self):
        """Return whether the client accepts the pagination preview media type"""
        accept_header = self.request.headers.get("Accept", "")
        if not accept_header:
            return False
        accepts = {s.strip().lower() for s in accept_header.strip().split(",")}
        return PAGINATION_MEDIA_TYPE in accepts

    def check_referer(self):
        """DEPRECATED"""
        warnings.warn(
            "check_referer is deprecated in JupyterHub 3.2 and always returns True",
            DeprecationWarning,
            stacklevel=2,
        )
        return True

    def check_post_content_type(self):
        """Check request content-type, e.g. for cross-site POST requests

        Cross-site POST via form will include content-type
        """
        content_type = self.request.headers.get("Content-Type")
        if not content_type:
            # not specified, e.g. from a script
            return True

        # parse content type for application/json
        fields = content_type.lower().split(";")
        if not any(f.lstrip().startswith("application/json") for f in fields):
            self.log.warning(f"Not allowing POST with content-type: {content_type}")
            return False

        return True

    async def prepare(self):
        await super().prepare()
        # tornado only checks xsrf on non-GET
        # we also check xsrf on GETs to API endpoints
        # make sure this runs after auth, which happens in super().prepare()
        if self.request.method not in {"HEAD", "OPTIONS"} and self.settings.get("xsrf_cookies"):
            self.check_xsrf_cookie()

    def check_xsrf_cookie(self):
        if getattr(self, "_token_authenticated", False):
            # if token-authenticated, ignore XSRF
            return
        return super().check_xsrf_cookie()

    def get_current_user_cookie(self):
        """Extend get_user_cookie to add checks for CORS"""
        cookie_user = super().get_current_user_cookie()
        # CORS checks for cookie-authentication
        # check these only if there is a cookie user,
        # avoiding misleading "Blocking Cross Origin" messages
        # when there's no cookie set anyway.
        if cookie_user:
            if self.request.method.upper() == "POST" and not self.check_post_content_type():
                return None
        return cookie_user

    def get_json_body(self):
        """Return the body of the request as JSON data."""
        if not self.request.body:
            return None
        body = self.request.body.strip().decode("utf-8")
        try:
            model = json.loads(body)
        except Exception:
            self.log.debug("Bad JSON: %r", body)
            self.log.error("Couldn't parse JSON", exc_info=True)
            raise web.HTTPError(400, reason="Invalid JSON in body of request")
        return model

    def write_error(self, status_code, **kwargs):
        """Write JSON errors instead of HTML"""
        exc_info = kwargs.get("exc_info")
        message = ""
        exception = None
        status_message = responses.get(status_code, "Unknown Error")
        if exc_info:
            exception = exc_info[1]
            # get the custom message, if defined
            try:
                message = exception.log_message % exception.args
            except Exception:
                pass

            # construct the custom reason, if defined
            reason = getattr(exception, "reason", "")
            if reason:
                status_message = reason

        if exception and isinstance(exception, SQLAlchemyError):
            try:
                exception_str = str(exception)
                self.log.warning("Rolling back session due to database error %s", exception_str)
            except Exception:
                self.log.warning("Rolling back session due to database error %s", type(exception))
            self.db.rollback()

        self.set_header("Content-Type", "application/json")
        if isinstance(exception, web.HTTPError):
            # allow setting headers from exceptions
            # since exception handler clears headers
            headers = getattr(exception, "headers", None)
            if headers:
                for key, value in headers.items():
                    self.set_header(key, value)
            # Content-Length must be recalculated.
            self.clear_header("Content-Length")

        self.write(json.dumps({"status": status_code, "message": message or status_message}))

    def token_model(self, token):
        """Get the JSON model for an APIToken"""

        owner_key = "user"
        owner = token.user.name

        model = {
            owner_key: owner,
            "id": token.api_id,
            "kind": "api_token",
            # deprecated field, but leave it present.
            "roles": [],
            "scopes": [],
            "created": isoformat(token.created),
            "last_activity": isoformat(token.last_activity),
            "expires_at": isoformat(token.expires_at),
            "note": token.note,
            "session_id": token.session_id,
            "oauth_client": token.oauth_client.description or token.oauth_client.identifier,
        }
        return model

    def _filter_model(self, model, access_map, entity, kind, keys=None):
        """
        Filter the model based on the available scopes and the entity requested for.
        If keys is a dictionary, update it with the allowed keys for the model.
        """
        allowed_keys = set()
        # for scope in access_map:
        #   scope_filter = self.get_scope_filter(scope)
        #   if scope_filter(entity, kind=kind):
        #       allowed_keys |= access_map[scope]
        model = {key: model[key] for key in allowed_keys if key in model}
        if isinstance(keys, set):
            keys.update(allowed_keys)
        return model

    _user_model_types = {
        "name": str,
        "admin": bool,
        "groups": list,
        "roles": list,
        "auth_state": dict,
    }

    _group_model_types = {"name": str, "users": list, "roles": list}

    _service_model_types = {
        "name": str,
        "admin": bool,
        "url": str,
        "oauth_client_allowed_scopes": list,
        "api_token": str,
        "info": dict,
        "display": bool,
        "oauth_no_confirm": bool,
        "command": list,
        "cwd": str,
        "environment": dict,
        "user": str,
        "oauth_client_id": str,
        "oauth_redirect_uri": str,
    }

    def _check_model(self, model, model_types, name):
        """Check a model provided by a REST API request

        Args:
            model (dict): user-provided model
            model_types (dict): dict of key:type used to validate types and keys
            name (str): name of the model, used in error messages
        """
        if not isinstance(model, dict):
            raise web.HTTPError(400, reason="Invalid JSON data: %r" % model)
        if not set(model).issubset(set(model_types)):
            raise web.HTTPError(400, reason="Invalid JSON keys: %r" % model)
        for key, value in model.items():
            if not isinstance(value, model_types[key]):
                raise web.HTTPError(
                    400, "%s.%s must be %s, not: %r" % (name, key, model_types[key], type(value))
                )

    def _check_user_model(self, model):
        """Check a request-provided user model from a REST API"""
        self._check_model(model, self._user_model_types, "user")
        for username in model.get("users", []):
            if not isinstance(username, str):
                raise web.HTTPError(400, ("usernames must be str, not %r", type(username)))

    def _check_group_model(self, model):
        """Check a request-provided group model from a REST API"""
        self._check_model(model, self._group_model_types, "group")
        for groupname in model.get("groups", []):
            if not isinstance(groupname, str):
                raise web.HTTPError(400, ("group names must be str, not %r", type(groupname)))

    def _check_service_model(self, model):
        """Check a request-provided service model from a REST API"""
        self._check_model(model, self._service_model_types, "service")
        service_name = model.get("name")
        if not isinstance(service_name, str):
            raise web.HTTPError(400, ("Service name must be str, not %r", type(service_name)))

    def get_api_pagination(self):
        default_limit = self.settings["api_page_default_limit"]
        max_limit = self.settings["api_page_max_limit"]
        if not self.accepts_pagination:
            # if new pagination Accept header is not used,
            # default to the higher max page limit to reduce likelihood
            # of missing users due to pagination in code that hasn't been updated
            default_limit = max_limit
        offset = self.get_argument("offset", None)
        limit = self.get_argument("limit", default_limit)
        try:
            offset = abs(int(offset)) if offset is not None else 0
            limit = abs(int(limit))
            if limit > max_limit:
                limit = max_limit
            if limit < 1:
                limit = 1
        except Exception:
            raise web.HTTPError(400, "Invalid argument type, offset and limit must be integers")
        return offset, limit

    def paginated_model(self, items, offset, limit, total_count):
        """Return the paginated form of a collection (list or dict)

        A dict with { items: [], _pagination: {}}
        instead of a single list (or dict).

        pagination info includes the current offset and limit,
        the total number of results for the query,
        and information about how to build the next page request
        if there is one.
        """
        next_offset = offset + limit
        data = {
            "items": items,
            "_pagination": {"offset": offset, "limit": limit, "total": total_count, "next": None},
        }
        if next_offset < total_count:
            # if there's a next page
            next_url_parsed = urlparse(self.request.full_url())
            query = parse_qs(next_url_parsed.query)
            query["offset"] = [next_offset]
            query["limit"] = [limit]
            next_url_parsed = next_url_parsed._replace(query=urlencode(query, doseq=True))
            next_url = urlunparse(next_url_parsed)
            data["_pagination"]["next"] = {"offset": next_offset, "limit": limit, "url": next_url}
        return data

    def options(self, *args, **kwargs):
        self.finish()


class API404(APIHandler):
    """404 for API requests

    Ensures JSON 404 errors for malformed URLs
    """

    def check_xsrf_cookie(self):
        pass

    async def prepare(self):
        await super().prepare()
        raise web.HTTPError(404)



from textwrap import dedent

from grader_service.convert.converters.autograde import AutogradeApp
from grader_service.convert.converters.baseapp import (
    ConverterApp,
    base_converter_aliases,
    base_converter_flags,
)
from grader_service.convert.converters.generate_assignment import GenerateAssignmentApp
from grader_service.convert.converters.generate_feedback import GenerateFeedbackApp

aliases = {}
aliases.update(base_converter_aliases)
aliases.update({})

flags = {}
flags.update(base_converter_flags)
flags.update({})


class GraderConverter(ConverterApp):
    name = "grader-converter"
    description = "Convert notebooks to different formats"
    version = ConverterApp.__version__

    aliases = aliases
    flags = flags

    subcommands = dict(
        generate_assignment=(
            GenerateAssignmentApp,
            dedent(
                """
                Create the student version of an assignment. Intended for use by
                instructors only.
                """
            ).strip(),
        ),
        autograde=(
            AutogradeApp,
            dedent(
                """
                Autograde submitted assignments. Intended for use by instructors
                only.
                """
            ).strip(),
        ),
        generate_feedback=(
            GenerateFeedbackApp,
            dedent(
                """
                Generate feedback (after autograding and manual grading).
                Intended for use by instructors only.
                """
            ).strip(),
        ),
    )


def main():
    GraderConverter.launch_instance()


if __name__ == "__main__":
    main()



import os
from textwrap import dedent
from typing import Any

from traitlets import Bool, List, default
from traitlets.config import Config

from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.convert import utils
from grader_service.convert.converters.base import BaseConverter
from grader_service.convert.converters.baseapp import ConverterApp
from grader_service.convert.preprocessors import (
    AddRevert,
    CheckCellMetadata,
    ClearAlwaysHiddenTests,
    ClearHiddenTests,
    ClearMarkScheme,
    ClearOutput,
    ClearSolutions,
    ComputeChecksums,
    IncludeHeaderFooter,
    LockCells,
    SaveCells,
)


class GenerateAssignment(BaseConverter):
    """Convert between the 'source code' and the 'student version' of assignment files."""

    create_assignment = Bool(
        True,
        help=dedent(
            """
            Whether to create the assignment at runtime if it does not
            already exist.
            """
        ),
    ).tag(config=True)

    @default("permissions")
    def _permissions_default(self) -> int:
        return 664

    preprocessors = List(
        [
            IncludeHeaderFooter,
            LockCells,
            ClearSolutions,
            ClearOutput,
            CheckCellMetadata,
            ComputeChecksums,
            SaveCells,
            ClearHiddenTests,
            ClearAlwaysHiddenTests,
            ClearMarkScheme,
            ComputeChecksums,
            AddRevert,
            CheckCellMetadata,
        ]
    ).tag(config=True)

    # NB: ClearHiddenTests must come after ComputeChecksums and SaveCells.
    # ComputerChecksums must come again after ClearHiddenTests.

    def __init__(
        self,
        input_dir: str,
        output_dir: str,
        file_pattern: str,
        assignment_settings: AssignmentSettings,
        **kwargs: Any,
    ) -> None:
        super().__init__(input_dir, output_dir, file_pattern, assignment_settings, **kwargs)
        self.force = True  # always overwrite generated assignments

        # load custom configuration if it exists
        c = Config()
        custom_config_path = f"{self._input_directory}/grader_config.py"
        if os.path.exists(custom_config_path):
            local_vars = {"c": c}
            with open(custom_config_path) as f:
                code = compile(f.read(), "config.py", "exec")
                exec(code, {}, local_vars)
            c = local_vars["c"]
        c.Exporter.default_preprocessors = []
        self.update_config(c)


class GenerateAssignmentApp(ConverterApp):
    version = ConverterApp.__version__

    def start(self):
        GenerateAssignment(
            input_dir=self.input_directory,
            output_dir=self.output_directory,
            file_pattern=self.file_pattern,
            assignment_settings=utils.get_assignment_settings_from_env(),
            config=self.config,
        ).start()



"""remove-group-table

Revision ID: f1ae66d52ad9
Revises: 28500016a3c3
Create Date: 2025-07-14 18:45:41.948031

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f1ae66d52ad9"
down_revision = "28500016a3c3"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table("partof")
    op.drop_table("group")


def downgrade():
    op.create_table(
        "group",
        sa.Column("id", sa.Integer(), autoincrement=True),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("lectid", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["lectid"], ["lecture.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "partof",
        sa.Column("username", sa.String(length=255), nullable=False),
        sa.Column("group_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["group_id"], ["group.id"]),
        sa.ForeignKeyConstraint(["username"], ["user.name"]),
        sa.PrimaryKeyConstraint("username", "group_id"),
    )



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

"""initial

Revision ID: a1791b1371ed
Revises:
Create Date: 2021-05-05 11:42:24.126371

"""

from datetime import UTC, datetime

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql.schema import UniqueConstraint

# revision identifiers, used by Alembic.
revision = "a1791b1371ed"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "lecture",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("code", sa.String(length=255), nullable=True, unique=True),
        sa.Column("state", sa.Enum("active", "complete", name="lecture_state"), nullable=False),
        sa.Column(
            "deleted",
            sa.Enum("active", "deleted", name="deleted"),
            server_default="active",
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), default=datetime.now(UTC), nullable=False),
        sa.Column(
            "updated_at",
            sa.DateTime(),
            default=datetime.now(UTC),
            onupdate=datetime.now(UTC),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
    )

    op.create_table(
        "user",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint("name"),
    )

    op.create_table(
        "assignment",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column(
            "type",
            sa.Enum("user", "group", name="assignment_type"),
            nullable=False,
            server_default="user",
        ),
        sa.Column("lectid", sa.Integer(), nullable=True),
        sa.Column("duedate", sa.DateTime(), nullable=True),
        sa.Column(
            "automatic_grading",
            sa.Enum("unassisted", "auto", "full_auto", name="automatic_grading"),
            server_default="unassisted",
            nullable=False,
        ),
        sa.Column("points", sa.DECIMAL(10, 3), nullable=False),
        sa.Column(
            "status",
            sa.Enum(
                "created",
                "pushed",
                "released",
                "fetching",
                "fetched",
                "complete",
                name="assignment_status",
            ),
            nullable=True,
        ),
        sa.Column(
            "deleted",
            sa.Enum("active", "deleted", name="deleted"),
            server_default="active",
            nullable=False,
        ),
        sa.Column("max_submissions", sa.Integer(), nullable=True, default=None, unique=False),
        sa.Column("allow_files", sa.Boolean(), nullable=False, default=False),
        sa.Column("properties", sa.Text(), nullable=True, unique=False),
        sa.Column("created_at", sa.DateTime(), default=datetime.now(UTC), nullable=False),
        sa.Column(
            "updated_at",
            sa.DateTime(),
            default=datetime.now(UTC),
            onupdate=datetime.now(UTC),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(["lectid"], ["lecture.id"]),
        sa.PrimaryKeyConstraint("id"),
        UniqueConstraint("name", "lectid", "deleted", name="u_name_in_lect"),
    )

    op.create_table(
        "takepart",
        sa.Column("username", sa.String(length=255), nullable=False),
        sa.Column("lectid", sa.Integer(), nullable=False),
        sa.Column("role", sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(["lectid"], ["lecture.id"]),
        sa.ForeignKeyConstraint(["username"], ["user.name"]),
        sa.PrimaryKeyConstraint("username", "lectid"),
    )

    op.create_table(
        "group",
        sa.Column("id", sa.Integer(), autoincrement=True),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("lectid", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["lectid"], ["lecture.id"]),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "partof",
        sa.Column("username", sa.String(length=255), nullable=False),
        sa.Column("group_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["group_id"], ["group.id"]),
        sa.ForeignKeyConstraint(["username"], ["user.name"]),
        sa.PrimaryKeyConstraint("username", "group_id"),
    )

    op.create_table(
        "submission",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("date", sa.DateTime(), nullable=True),
        sa.Column(
            "auto_status",
            sa.Enum(
                "pending",
                "not_graded",
                "automatically_graded",
                "grading_failed",
                name="auto_status",
            ),
            nullable=False,
        ),
        sa.Column(
            "manual_status",
            sa.Enum("not_graded", "manually_graded", "being_edited", name="manual_status"),
            nullable=False,
        ),
        sa.Column("edited", sa.Boolean(), default=False),
        sa.Column("score", sa.DECIMAL(10, 3), nullable=True),
        sa.Column("assignid", sa.Integer(), nullable=True),
        sa.Column("username", sa.String(length=255), nullable=True),
        sa.Column("commit_hash", sa.String(length=40), nullable=False),
        sa.Column("feedback_available", sa.Boolean(), nullable=False, server_default="false"),
        sa.Column(
            "updated_at",
            sa.DateTime(),
            default=datetime.now(UTC),
            onupdate=datetime.now(UTC),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(["assignid"], ["assignment.id"]),
        sa.ForeignKeyConstraint(["username"], ["user.name"]),
        sa.PrimaryKeyConstraint("id"),
    )

    op.create_table(
        "submission_logs",
        sa.Column("sub_id", sa.Integer(), nullable=False),
        sa.Column("logs", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(["sub_id"], ["submission.id"]),
        sa.PrimaryKeyConstraint("sub_id"),
    )

    op.create_table(
        "submission_properties",
        sa.Column("sub_id", sa.Integer(), nullable=False),
        sa.Column("properties", sa.Text(), nullable=True, unique=False),
        sa.ForeignKeyConstraint(["sub_id"], ["submission.id"]),
        sa.PrimaryKeyConstraint("sub_id"),
    )

    # ### end Alembic commands ###


def downgrade():
    # This does not drop enum types which are required when using postgresql
    op.drop_table("submission_logs")
    op.drop_table("submission_properties")
    op.drop_table("submission")
    op.drop_table("takepart")
    op.drop_table("assignment")
    op.drop_table("partof")
    op.drop_table("group")
    op.drop_table("lecture")
    op.drop_table("user")

    # drop enums
    if sa.inspect(op.get_bind()).dialect.name == "postgresql":
        op.execute("DROP TYPE IF EXISTS lecture_state")
        op.execute("DROP TYPE IF EXISTS assignment_type")
        op.execute("DROP TYPE IF EXISTS automatic_grading")
        op.execute("DROP TYPE IF EXISTS assignment_status")
        op.execute("DROP TYPE IF EXISTS deleted")
        op.execute("DROP TYPE IF EXISTS auto_status")
        op.execute("DROP TYPE IF EXISTS manual_status")
    # ### end Alembic commands ###



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import os
from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool

import grader_service.orm

# Alembic configuration
config = context.config

# Set up logging from the config file
fileConfig(config.config_file_name)


def configure_db_url():
    """Set the database URL from environment variables if not set in config."""
    if not config.get_main_option("sqlalchemy.url"):
        db_url = os.getenv("GRADER_DB_URL", "sqlite:///grader.db")
        config.set_main_option("sqlalchemy.url", db_url)


configure_db_url()

# Metadata for autogenerate support
target_metadata = grader_service.orm.Base.metadata


def run_migrations_offline():
    """Run migrations *without* a SQL connection."""
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
        render_as_batch=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations *with* a SQL connection."""
    # Retrieve the existing connection if it has already been initialized
    connectable = config.attributes.get("connection", None)

    if connectable is None:
        # No existing connection, so we create a new one
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
            future=True,
        )

        # New connection case: Begin a transaction and run migrations
        with connectable.connect() as connection:
            # Configure the Alembic context with the new connection
            context.configure(
                connection=connection,
                target_metadata=target_metadata,
                render_as_batch=True,  # Needed for SQLite or other databases
            )

            # Run migrations within a transaction
            with context.begin_transaction():
                context.run_migrations()
    else:
        # Existing connection case: Simply run migrations without reconnecting
        # Configure the Alembic context with the provided connection
        context.configure(
            connection=connectable, target_metadata=target_metadata, render_as_batch=True
        )

        # Run migrations within a transaction
        with context.begin_transaction():
            context.run_migrations()


# Run migrations according to the mode
if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# flake8: noqa
# import models into model package
from grader_service.api.models.assignment import Assignment
from grader_service.api.models.assignment_detail import AssignmentDetail
from grader_service.api.models.assignment_settings import AssignmentSettings
from grader_service.api.models.auto_status import AutoStatus
from grader_service.api.models.error_message import ErrorMessage
from grader_service.api.models.feedback_status import FeedbackStatus
from grader_service.api.models.lecture import Lecture
from grader_service.api.models.manual_status import ManualStatus
from grader_service.api.models.remote_file_status import RemoteFileStatus
from grader_service.api.models.submission import Submission
from grader_service.api.models.submission_period import SubmissionPeriod
from grader_service.api.models.user import User
from grader_service.api.models.user_submissions_inner import UserSubmissionsInner



import json
from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode

from grader_service.convert.gradebook.gradebook import Gradebook
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class OverwriteKernelspec(NbGraderPreprocessor):
    """A preprocessor for checking the notebook kernelspec metadata."""

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # pull information from the resources
        self.notebook_id = resources["unique_key"]
        self.json_path = resources["output_json_path"]

        with Gradebook(self.json_path) as gb:
            kernelspec = json.loads(gb.find_notebook(self.notebook_id).kernelspec)
            self.log.debug("Source notebook kernelspec: {}".format(kernelspec))
            self.log.debug(
                "Submitted notebook kernelspec: {}".format(nb.metadata.get("kernelspec", None))
            )
            if kernelspec:
                self.log.info("Overwriting submitted notebook kernelspec: {}".format(kernelspec))
                nb.metadata["kernelspec"] = kernelspec
        return nb, resources



"""add_submission_softdelete_column

Revision ID: a0718dae969d
Revises: 9cfeb0faa0c0
Create Date: 2024-07-06 12:12:59.091247

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a0718dae969d"
down_revision = "ba71c755c153"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "submission",
        sa.Column(
            "deleted",
            sa.Enum("active", "deleted", name="deleted"),
            server_default="active",
            nullable=False,
        ),
    )

    op.execute("UPDATE submission SET deleted = 'active'")


def downgrade():
    op.drop_column("submission", "deleted")



"""All custom exceptions defined in one place."""


class ValidationError(Exception):
    """Base class for validation exceptions."""


class MissingRequiredArgumentError(ValidationError):
    """Exception raised for missing required request arguments."""

    pass


class IncorrectValueError(ValidationError):
    """Exception raised for incorrect values."""

    pass


class TokenError(ValidationError):
    """Exception raised for failed JWT decoding or verification."""

    pass


class InvalidAudienceError(ValidationError):
    """Exception raised for invalid audience."""

    pass


class LoginError(Exception):
    """Lookup of username in ID token failed"""

    pass



import os
from typing import Optional

from grader_service.auth.auth import Authenticator
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.handlers.base_handler import BaseHandler
from grader_service.orm import User, Lecture
from grader_service.orm.base import DeleteState
from grader_service.orm.lecture import LectureState
from grader_service.orm.takepart import Scope, Role
from traitlets import log as traitlets_log


logger = traitlets_log.get_logger()

logger.info("### loading service config")

c.GraderService.service_host = "127.0.0.1"
# existing directory to use as the base directory for the grader service
service_dir = os.path.join(os.getcwd(), "service_dir")
c.GraderService.grader_service_dir = service_dir

c.RequestHandlerConfig.autograde_executor_class = LocalAutogradeExecutor

c.CeleryApp.conf = dict(
    broker_url="amqp://localhost",
    result_backend="rpc://",
    task_serializer="json",
    result_serializer="json",
    accept_content=["json"],
    broker_connection_retry_on_startup=True,
    task_always_eager=True,
)
c.CeleryApp.worker_kwargs = dict(concurrency=1, pool="prefork")


# JupyterHub client config
c.GraderService.oauth_clients = [
    {
        "client_id": "my_id",
        "client_secret": "my_secret",
        "redirect_uri": "http://localhost:8080/hub/oauth_callback",
    }
]

from grader_service.auth.token import JupyterHubTokenAuthenticator

c.GraderService.authenticator_class = JupyterHubTokenAuthenticator

c.JupyterHubTokenAuthenticator.user_info_url = "http://localhost:8080/hub/api/user"


def post_auth_hook(authenticator: Authenticator, handler: BaseHandler, authentication: dict):
    log = handler.log
    log.info("post_auth_hook started")

    session = handler.session
    groups: list[str] = authentication["groups"]

    username = authentication["name"]
    user_model: Optional[User] = session.query(User).filter(User.name == username).one_or_none()
    if user_model is None:
        user_model = User()
        user_model.name = username
        user_model.display_name = username
        session.add(user_model)
        session.commit()

    for group in groups:
        if ":" in group:
            split_group = group.split(":")
            lecture_code = split_group[0]
            scope = split_group[1]
            scope = Scope[scope]

            lecture = session.query(Lecture).filter(Lecture.code == lecture_code).one_or_none()
            if lecture is None:
                lecture = Lecture()
                lecture.code = lecture_code
                lecture.name = lecture_code
                lecture.state = LectureState.active
                lecture.deleted = DeleteState.active
                session.add(lecture)
                session.commit()

            role = (
                session.query(Role)
                .filter(Role.user_id == user_model.id, Role.lectid == lecture.id)
                .one_or_none()
            )
            if role is None:
                log.info(
                    "No role for user %s in lecture %s... creating role", username, lecture_code
                )
                role = Role(user_id=user_model.id, lectid=lecture.id, role=scope)
                session.add(role)
                session.commit()
            else:
                log.info(
                    "Found role %s for user %s in lecture %s... updating role to %s",
                    role.role.name,
                    username,
                    lecture_code,
                    scope.name,
                )
                role.role = scope
                session.commit()
        else:
            log.info("Found group that doesn't match schema. Ignoring %s", group)

    return authentication


c.Authenticator.post_auth_hook = post_auth_hook

c.Authenticator.allowed_users = {"admin", "instructor", "student", "tutor"}

c.Authenticator.admin_users = {"admin"}

c.GraderService.load_roles = {
    "lecture1": [
        {"members": ["student"], "role": "student"},
        {"members": ["tutor"], "role": "tutor"},
        {"members": ["instructor", "admin"], "role": "instructor"},
    ]
}



from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api.models.submission_period import SubmissionPeriod
from grader_service.api import util

from grader_service.api.models.submission_period import SubmissionPeriod  # noqa: E501

class AssignmentSettings(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, deadline=None, max_submissions=None, allowed_files=[], late_submission=None, autograde_type='auto', group=None, cell_timeout=None):  # noqa: E501
        """AssignmentSettings - a model defined in OpenAPI

        :param deadline: The deadline of this AssignmentSettings.  # noqa: E501
        :type deadline: datetime
        :param max_submissions: The max_submissions of this AssignmentSettings.  # noqa: E501
        :type max_submissions: int
        :param allowed_files: The allowed_files of this AssignmentSettings.  # noqa: E501
        :type allowed_files: List[str]
        :param late_submission: The late_submission of this AssignmentSettings.  # noqa: E501
        :type late_submission: List[SubmissionPeriod]
        :param autograde_type: The autograde_type of this AssignmentSettings.  # noqa: E501
        :type autograde_type: str
        :param group: The group of this AssignmentSettings.  # noqa: E501
        :type group: str
        :param cell_timeout: The cell_timeout of this AssignmentSettings.  # noqa: E501
        :type cell_timeout: int
        """
        self.openapi_types = {
            'deadline': datetime,
            'max_submissions': int,
            'allowed_files': List[str],
            'late_submission': List[SubmissionPeriod],
            'autograde_type': str,
            'group': str,
            'cell_timeout': int
        }

        self.attribute_map = {
            'deadline': 'deadline',
            'max_submissions': 'max_submissions',
            'allowed_files': 'allowed_files',
            'late_submission': 'late_submission',
            'autograde_type': 'autograde_type',
            'group': 'group',
            'cell_timeout': 'cell_timeout'
        }

        self._deadline = deadline
        self._max_submissions = max_submissions
        self._allowed_files = allowed_files
        self._late_submission = late_submission
        self._autograde_type = autograde_type
        self._group = group
        self._cell_timeout = cell_timeout

    @classmethod
    def from_dict(cls, dikt) -> 'AssignmentSettings':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The AssignmentSettings of this AssignmentSettings.  # noqa: E501
        :rtype: AssignmentSettings
        """
        return util.deserialize_model(dikt, cls)

    @property
    def deadline(self) -> datetime:
        """Gets the deadline of this AssignmentSettings.


        :return: The deadline of this AssignmentSettings.
        :rtype: datetime
        """
        return self._deadline

    @deadline.setter
    def deadline(self, deadline: datetime):
        """Sets the deadline of this AssignmentSettings.


        :param deadline: The deadline of this AssignmentSettings.
        :type deadline: datetime
        """

        self._deadline = deadline

    @property
    def max_submissions(self) -> int:
        """Gets the max_submissions of this AssignmentSettings.


        :return: The max_submissions of this AssignmentSettings.
        :rtype: int
        """
        return self._max_submissions

    @max_submissions.setter
    def max_submissions(self, max_submissions: int):
        """Sets the max_submissions of this AssignmentSettings.


        :param max_submissions: The max_submissions of this AssignmentSettings.
        :type max_submissions: int
        """

        self._max_submissions = max_submissions

    @property
    def allowed_files(self) -> List[str]:
        """Gets the allowed_files of this AssignmentSettings.


        :return: The allowed_files of this AssignmentSettings.
        :rtype: List[str]
        """
        return self._allowed_files

    @allowed_files.setter
    def allowed_files(self, allowed_files: List[str]):
        """Sets the allowed_files of this AssignmentSettings.


        :param allowed_files: The allowed_files of this AssignmentSettings.
        :type allowed_files: List[str]
        """

        self._allowed_files = allowed_files

    @property
    def late_submission(self) -> List[SubmissionPeriod]:
        """Gets the late_submission of this AssignmentSettings.


        :return: The late_submission of this AssignmentSettings.
        :rtype: List[SubmissionPeriod]
        """
        return self._late_submission

    @late_submission.setter
    def late_submission(self, late_submission: List[SubmissionPeriod]):
        """Sets the late_submission of this AssignmentSettings.


        :param late_submission: The late_submission of this AssignmentSettings.
        :type late_submission: List[SubmissionPeriod]
        """

        self._late_submission = late_submission

    @property
    def autograde_type(self) -> str:
        """Gets the autograde_type of this AssignmentSettings.


        :return: The autograde_type of this AssignmentSettings.
        :rtype: str
        """
        return self._autograde_type

    @autograde_type.setter
    def autograde_type(self, autograde_type: str):
        """Sets the autograde_type of this AssignmentSettings.


        :param autograde_type: The autograde_type of this AssignmentSettings.
        :type autograde_type: str
        """
        allowed_values = ["auto", "full_auto", "unassisted"]  # noqa: E501
        if autograde_type not in allowed_values:
            raise ValueError(
                "Invalid value for `autograde_type` ({0}), must be one of {1}"
                .format(autograde_type, allowed_values)
            )

        self._autograde_type = autograde_type

    @property
    def group(self) -> str:
        """Gets the group of this AssignmentSettings.


        :return: The group of this AssignmentSettings.
        :rtype: str
        """
        return self._group

    @group.setter
    def group(self, group: str):
        """Sets the group of this AssignmentSettings.


        :param group: The group of this AssignmentSettings.
        :type group: str
        """

        self._group = group

    @property
    def cell_timeout(self) -> int:
        """Gets the cell_timeout of this AssignmentSettings.


        :return: The cell_timeout of this AssignmentSettings.
        :rtype: int
        """
        return self._cell_timeout

    @cell_timeout.setter
    def cell_timeout(self, cell_timeout: int):
        """Sets the cell_timeout of this AssignmentSettings.


        :param cell_timeout: The cell_timeout of this AssignmentSettings.
        :type cell_timeout: int
        """

        self._cell_timeout = cell_timeout



"""Added oauth provider functionality

Revision ID: ba71c755c153
Revises: 9cfeb0faa0c0
Create Date: 2023-08-18 10:30:24.465356

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import (
    Column,
    DateTime,
    ForeignKeyConstraint,
    Integer,
    PrimaryKeyConstraint,
    Text,
    Unicode,
)

from grader_service.utils import new_token

# revision identifiers, used by Alembic.
revision = "ba71c755c153"
down_revision = "9cfeb0faa0c0"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()
    if connection.dialect.name == "sqlite":
        # sqlite has to recreate the tables on `batch_alter_table`, but dropping a table
        # would cause integrity errors, so we disable the foreign key constraint temporarily
        connection.execute(sa.text("PRAGMA foreign_keys=OFF"))

    op.create_table(
        "api_token",
        Column("username", Unicode(255)),
        Column("id", Integer, nullable=True),
        Column("hashed", Unicode(255)),
        Column("prefix", Unicode(16)),
        Column("client_id", Unicode(255)),
        Column("session_id", Unicode(255)),
        Column("created", DateTime),
        Column("expires_at", DateTime),
        Column("last_activity", DateTime),
        Column("note", Unicode(1023)),
        Column("scopes", Text),
        PrimaryKeyConstraint("id"),
        ForeignKeyConstraint(["username"], ["user.name"]),
    )

    op.create_table(
        "oauth_client",
        Column("id", Integer, nullable=True),
        Column("identifier", Unicode(255), unique=True),
        Column("description", Unicode(1023)),
        Column("secret", Unicode(255)),
        Column("redirect_uri", Unicode(1023)),
        Column("allowed_scopes", Text),
        PrimaryKeyConstraint("id"),
    )

    op.create_table(
        "oauth_code",
        Column("id", Integer, nullable=True),
        Column("client_id", Unicode(255)),
        Column("code", Unicode(36)),
        Column("expires_at", Integer),
        Column("redirect_uri", Unicode(1023)),
        Column("session_id", Unicode(255)),
        Column("username", Unicode(255)),
        Column("scopes", Text),
        PrimaryKeyConstraint("id"),
        ForeignKeyConstraint(["client_id"], ["oauth_client.identifier"]),
        ForeignKeyConstraint(["username"], ["user.name"]),
    )

    op.add_column("user", sa.Column("encrypted_auth_state", sa.types.LargeBinary, nullable=True))
    op.add_column("user", sa.Column("cookie_id", Unicode(255), nullable=True))

    result = connection.execute(sa.text('select * from "user"')).mappings()
    for row in result:
        connection.execute(
            sa.text('UPDATE "user" SET cookie_id = :uuid WHERE name = :name'),
            {"uuid": new_token(), "name": row["name"]},
        )

    if connection.dialect.name != "sqlite":
        op.alter_column("user", "cookie_id", nullable=False)
        op.create_unique_constraint("uq_user_cookie", "user", ["cookie_id"])
    else:
        with op.batch_alter_table("user") as batch_op:
            batch_op.alter_column("cookie_id", nullable=False)
            batch_op.create_unique_constraint("uq_user_cookie", ["cookie_id"])


def downgrade():
    dialect = op.get_bind().dialect.name
    if dialect == "sqlite":
        # sqlite has to recreate the tables on `batch_alter_table`, but dropping a table
        # would cause integrity errors, so we disable the foreign key constraint temporarily
        op.execute(sa.text("PRAGMA foreign_keys=OFF"))

    op.drop_table("api_token")
    op.drop_table("oauth_code")
    op.drop_table("oauth_client")
    op.drop_column("user", "encrypted_auth_state")
    with op.batch_alter_table("user") as batch_op:
        batch_op.drop_constraint("uq_user_cookie")

    op.drop_column("user", "cookie_id")



from textwrap import dedent
from typing import Any, Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from traitlets import Bool, Unicode

from grader_service.convert import utils
from grader_service.convert.gradebook.gradebook import Gradebook, MissingEntry
from grader_service.convert.nbgraderformat import MetadataValidator
from grader_service.convert.preprocessors.base import NbGraderPreprocessor


class OverwriteCells(NbGraderPreprocessor):
    """A preprocessor to overwrite information about grade and solution cells."""

    add_missing_cells = Bool(
        True,
        help=dedent(
            """
            Whether or not missing grade_cells should be added back
            to the notebooks being graded.
            """
        ),
    ).tag(config=True)

    missing_cell_notification = Unicode(
        "This cell (id:{cell_id}) was missing from the submission. "
        + "It was added back by the grader.\n\n",  # markdown requires two newlines
        help=dedent(
            """
            A text to add at the beginning of every missing cell re-added to the notebook during autograding.
            """
        ),
    ).tag(config=True)

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # pull information from the resources
        self.notebook_id = resources["unique_key"]
        self.json_path = resources["output_json_path"]
        self.gradebook = Gradebook(self.json_path)

        with self.gradebook:
            nb, resources = super(OverwriteCells, self).preprocess(nb, resources)
            if self.add_missing_cells:
                nb, resources = self.add_missing_grade_cells(nb, resources)
                nb, resources = self.add_missing_task_cells(nb, resources)

        return nb, resources

    def missing_cell_transform(self, source_cell, max_score, is_solution=False, is_task=False):
        """
        Converts source_cell obtained from Gradebook into a cell that can be added to the notebook.
        It is assumed that the cell is a grade_cell (unless is_task=True)
        """

        missing_cell_notification = self.missing_cell_notification.format(cell_id=source_cell.name)

        cell = {
            "cell_type": source_cell.cell_type,
            "metadata": {
                "deletable": False,
                "editable": False,
                "nbgrader": {
                    "grade": True,
                    "grade_id": source_cell.name,
                    "locked": source_cell.locked,
                    "checksum": source_cell.checksum,
                    "cell_type": source_cell.cell_type,
                    "points": max_score,
                    "solution": False,
                },
            },
            "source": missing_cell_notification + source_cell.source,
        }

        # Code cell format is slightly different
        if cell["cell_type"] == "code":
            cell["execution_count"] = None
            cell["outputs"] = []
            cell["source"] = "# " + cell["source"]  # make the notification we add a comment

        # some grade cells are editable (manually graded answers)
        if is_solution:
            del cell["metadata"]["editable"]
            cell["metadata"]["nbgrader"]["solution"] = True
        # task cells are also a bit different
        elif is_task:
            cell["metadata"]["nbgrader"]["grade"] = False
            cell["metadata"]["nbgrader"]["task"] = True
            # this is when task cells were added so metadata validation should start from here
            cell["metadata"]["nbgrader"]["schema_version"] = 3

        cell = NotebookNode(cell)
        cell = MetadataValidator().upgrade_cell_metadata(cell)
        return cell

    def add_missing_grade_cells(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        """
        Add missing grade cells back to the notebook.
        If missing, find the previous solution/grade cell, and add the current cell after it.
        It is assumed such a cell exists because
        presumably the grade_cell exists to grade some work in the solution cell.
        """
        source_nb = self.gradebook.find_notebook(self.notebook_id)
        source_cells = source_nb.source_cells
        source_cell_ids = [cell.name for cell in source_cells]
        grade_cells = {cell.name: cell for cell in source_nb.grade_cells}
        solution_cell_ids = [cell.name for cell in source_nb.solution_cells]

        # track indices of solution and grade cells in the submitted notebook
        submitted_cell_idxs = dict()
        for idx, cell in enumerate(nb.cells):
            if utils.is_grade(cell) or utils.is_solution(cell):
                submitted_cell_idxs[cell.metadata.nbgrader["grade_id"]] = idx

        # Every time we add a cell, the idxs above get shifted
        # We could process the notebook backwards, but that makes adding the cells in the right place more difficult
        # So we keep track of how many we have added so far
        added_count = 0

        for grade_cell_id, grade_cell in grade_cells.items():
            # If missing, find the previous solution/grade cell, and add the current cell after it.
            if grade_cell_id not in submitted_cell_idxs:
                self.log.warning(
                    f"Missing grade cell {grade_cell_id} encountered, adding to notebook"
                )
                source_cell_idx = source_cell_ids.index(grade_cell_id)
                cell_to_add = source_cells[source_cell_idx]
                cell_to_add = self.missing_cell_transform(
                    cell_to_add,
                    grade_cell.max_score,
                    is_solution=grade_cell_id in solution_cell_ids,
                )
                # First cell was deleted, add it to start
                if source_cell_idx == 0:
                    nb.cells.insert(0, cell_to_add)
                    submitted_cell_idxs[grade_cell_id] = 0
                # Deleted cell is not the first, add it after the previous solution/grade cell
                else:
                    prev_cell_id = source_cell_ids[source_cell_idx - 1]
                    prev_cell_idx = submitted_cell_idxs[prev_cell_id] + added_count
                    nb.cells.insert(prev_cell_idx + 1, cell_to_add)  # +1 to add it after
                    submitted_cell_idxs[grade_cell_id] = submitted_cell_idxs[prev_cell_id]

                # If the cell we just added is followed by other missing cells, we need to know its index in the nb
                # However, no need to add `added_count` to avoid double-counting

                added_count += 1  # shift idxs

        return nb, resources

    def add_missing_task_cells(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        """
        Add missing task cells back to the notebook.
        We can't figure out their original location, so they are added at the end, in their original order.
        """
        source_nb = self.gradebook.find_notebook(self.notebook_id)
        source_cells = source_nb.source_cells
        source_cell_ids = [cell.name for cell in source_cells]
        submitted_ids = [
            cell["metadata"]["nbgrader"]["grade_id"]
            for cell in nb.cells
            if "nbgrader" in cell["metadata"]
        ]
        for task_cell in source_nb.task_cells:
            if task_cell.name not in submitted_ids:
                cell_to_add = source_cells[source_cell_ids.index(task_cell.name)]
                cell_to_add = self.missing_cell_transform(
                    cell_to_add, task_cell.max_score, is_task=True
                )
                nb.cells.append(cell_to_add)

        return nb, resources

    def add_missing_readonly_cells(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        """
        Add missing task cells back to the notebook.
        We can't figure out their original location, so they are added at the end, in their original order.
        """
        source_nb = self.gradebook.find_notebook(self.notebook_id)
        source_cells = source_nb.source_cells
        source_cell_ids = [cell.name for cell in source_cells]
        submitted_ids = [
            cell["metadata"]["nbgrader"]["grade_id"]
            for cell in nb.cells
            if "nbgrader" in cell["metadata"]
        ]
        for task_cell in source_nb.task_cells:
            if task_cell.name not in submitted_ids:
                cell_to_add = source_cells[source_cell_ids.index(task_cell.name)]
                cell_to_add = self.missing_cell_transform(
                    cell_to_add, task_cell.max_score, is_task=True
                )
                nb.cells.insert(cell_to_add, 0)

        return nb, resources

    def update_cell_type(self, cell: NotebookNode, cell_type: str) -> None:
        if cell.cell_type == cell_type:
            return
        elif cell_type == "code":
            cell.cell_type = "code"
            cell.outputs = []
            cell.execution_count = None

        elif cell_type == "markdown":
            cell.cell_type = "markdown"
            if "outputs" in cell:
                del cell["outputs"]
            if "execution_count" in cell:
                del cell["execution_count"]
        elif cell_type == "raw":
            cell.cell_type = "raw"

    def report_change(self, name: str, attr: str, old: Any, new: Any) -> None:
        self.log.warning(
            "Attribute '%s' for cell %s has changed! (should be: %s, got: %s)", attr, name, old, new
        )

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        grade_id = cell.metadata.get("nbgrader", {}).get("grade_id", None)
        if grade_id is None:
            return cell, resources

        try:
            source_cell = self.gradebook.find_source_cell(grade_id, self.notebook_id)
        except MissingEntry:
            self.log.warning("Cell '{}' does not exist in the properties".format(grade_id))
            del cell.metadata.nbgrader["grade_id"]
            return cell, resources

        # check that the cell type hasn't changed
        if cell.cell_type != source_cell.cell_type:
            self.report_change(grade_id, "cell_type", source_cell.cell_type, cell.cell_type)
            self.update_cell_type(cell, source_cell.cell_type)

        # check that the locked status hasn't changed
        if utils.is_locked(cell) != source_cell.locked:
            self.report_change(grade_id, "locked", source_cell.locked, utils.is_locked(cell))
            cell.metadata.nbgrader["locked"] = source_cell.locked

        # if it's a grade cell, check that the max score hasn't changed
        if utils.is_grade(cell):
            grade_cell = self.gradebook.find_graded_cell(grade_id, self.notebook_id)
            old_points = float(grade_cell.max_score)
            new_points = float(cell.metadata.nbgrader["points"])

            if old_points != new_points:
                self.report_change(grade_id, "points", old_points, new_points)
                cell.metadata.nbgrader["points"] = old_points

        # always update the checksum, just in case
        cell.metadata.nbgrader["checksum"] = source_cell.checksum

        # if it's locked, check that the checksum hasn't changed
        if source_cell.locked:
            old_checksum = source_cell.checksum
            new_checksum = utils.compute_checksum(cell)
            if old_checksum != new_checksum:
                self.log.info(
                    "Source of cell {} in overwritten order to prove checksum.".format(grade_id)
                )
                cell.source = source_cell.source
                # check the checksum is correct now
                double_checksum = utils.compute_checksum(cell)
                if double_checksum != old_checksum:
                    self.log.error(
                        "Checksums of the cell {} and source cell {} are not the same.".format(
                            double_checksum, new_checksum
                        )
                    )
                    raise RuntimeError(
                        "Inconsistent checksums for cell {}".format(source_cell.name)
                    )

        return cell, resources



from http import HTTPStatus

from tornado.web import HTTPError

from grader_service.handlers.base_handler import GraderBaseHandler
from grader_service.orm.lecture import Lecture, LectureState
from grader_service.registry import VersionSpecifier, register_handler


@register_handler(path=r"\/api\/health\/?", version_specifier=VersionSpecifier.ALL)
class HealthHandler(GraderBaseHandler):
    """
    Tornado Handler class for http requests to /health.
    """

    async def get(self):
        """
        Check health of service
        :return permissions of the user
        """
        try:
            self.session.query(Lecture).filter(Lecture.state == LectureState.active).all()
        except Exception as e:
            self.log.error(e)
            raise HTTPError(HTTPStatus.INTERNAL_SERVER_ERROR, reason="Database Error")

        response = {"health": "OK"}
        self.write_json(response)



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import enum
from datetime import UTC, datetime

from sqlalchemy import Column, DateTime, Enum, Integer, String
from sqlalchemy.orm import relationship

from grader_service.api.models import lecture
from grader_service.orm.base import Base, DeleteState, Serializable


class LectureState(enum.IntEnum):
    active = 0
    complete = 1


class Lecture(Base, Serializable):
    __tablename__ = "lecture"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255), nullable=True, unique=False)
    code = Column(String(255), nullable=True, unique=True)
    state = Column(Enum(LectureState), nullable=False, unique=False)
    deleted = Column(Enum(DeleteState), nullable=False, unique=False)

    created_at = Column(DateTime, default=datetime.now(UTC), nullable=False)
    updated_at = Column(
        DateTime, default=datetime.now(UTC), onupdate=datetime.now(UTC), nullable=False
    )

    assignments = relationship(
        "Assignment", back_populates="lecture", cascade="all, delete-orphan", passive_deletes=True
    )
    roles = relationship(
        "Role", back_populates="lecture", cascade="all, delete-orphan", passive_deletes=True
    )

    @property
    def model(self) -> lecture.Lecture:
        return lecture.Lecture(
            id=self.id, name=self.name, code=self.code, complete=self.state == LectureState.complete
        )



from datetime import datetime, timedelta, timezone

from sqlalchemy import Column, DateTime, ForeignKey, Integer, Unicode, inspect, or_
from sqlalchemy.orm import relationship
from sqlalchemy.orm.strategy_options import joinedload
from tornado.log import app_log

from grader_service.orm.base import Base
from grader_service.orm.json_util import JSONList
from grader_service.utils import compare_token, hash_token, new_token, utcnow


class Expiring:
    """Mixin for expiring entries

    Subclass must define at least expires_at property,
    which should be unix timestamp or datetime object
    """

    now = utcnow  # function, must return float timestamp or datetime
    expires_at = None  # must be defined

    @property
    def expires_in(self):
        """Property returning expiration in seconds from now

        or None
        """
        if self.expires_at:
            delta = self.expires_at - self.now()
            if isinstance(delta, timedelta):
                delta = delta.total_seconds()
            return delta
        else:
            return None

    @classmethod
    def purge_expired(cls, db):
        """Purge expired API Tokens from the database"""
        now = cls.now()
        deleted = False
        for obj in db.query(cls).filter(cls.expires_at is not None).filter(cls.expires_at < now):
            app_log.debug("Purging expired %s", obj)
            deleted = True
            db.delete(obj)
        if deleted:
            db.commit()


class Hashed(Expiring):
    """Mixin for tables with hashed tokens"""

    prefix_length = 4
    algorithm = "sha512"
    rounds = 16384
    salt_bytes = 8
    min_length = 8

    # values to use for internally generated tokens,
    # which have good entropy as UUIDs
    generated = True
    generated_salt_bytes = 8
    generated_rounds = 1

    @property
    def token(self):
        raise AttributeError("token is write-only")

    @token.setter
    def token(self, token):
        """Store the hashed value and prefix for a token"""
        self.prefix = token[: self.prefix_length]
        if self.generated:
            # Generated tokens are UUIDs, which have sufficient entropy on their own
            # and don't need salt & hash rounds.
            # ref: https://security.stackexchange.com/a/151262/155114
            rounds = self.generated_rounds
            salt_bytes = self.generated_salt_bytes
        else:
            rounds = self.rounds
            salt_bytes = self.salt_bytes
        self.hashed = hash_token(token, rounds=rounds, salt=salt_bytes, algorithm=self.algorithm)

    def match(self, token):
        """Is this my token?"""
        return compare_token(self.hashed, token)

    @classmethod
    def check_token(cls, db, token):
        """Check if a token is acceptable"""
        if len(token) < cls.min_length:
            raise ValueError(
                "Tokens must be at least %i characters, got %r" % (cls.min_length, token)
            )
        found = cls.find(db, token)
        if found:
            raise ValueError("Collision on token: %s..." % token[: cls.prefix_length])

    @classmethod
    def find_prefix(cls, db, token):
        """Start the query for matching token.

        Returns an SQLAlchemy query already filtered by prefix-matches.

        .. versionchanged:: 1.2

            Excludes expired matches.
        """
        prefix = token[: cls.prefix_length]
        # since we can't filter on hashed values, filter on prefix
        # so we aren't comparing with all tokens
        prefix_match = db.query(cls).filter_by(prefix=prefix)
        prefix_match = prefix_match.filter(or_(cls.expires_at is None, cls.expires_at >= cls.now))
        return prefix_match

    @classmethod
    def find(cls, db, token):
        """Find a token object by value.

        Returns None if not found.

        `kind='user'` only returns API tokens for users
        """
        prefix_match = cls.find_prefix(db, token).options(joinedload(cls.user))

        for orm_token in prefix_match:
            if orm_token.match(token):
                return orm_token


class APIToken(Hashed, Base):
    """An API token"""

    __tablename__ = "api_token"

    user_id = Column(Integer, ForeignKey("user.id", ondelete="CASCADE"), nullable=True)

    user = relationship("User", back_populates="api_tokens")
    oauth_client = relationship("OAuthClient", back_populates="access_tokens")

    id = Column(Integer, primary_key=True)
    hashed = Column(Unicode(255), unique=True)
    prefix = Column(Unicode(16), index=True)

    @property
    def api_id(self):
        return "a%i" % self.id

    @property
    def owner(self):
        return self.user

    client_id = Column(
        Unicode(255),
        ForeignKey("oauth_client.identifier", ondelete="CASCADE"),
        nullable=True,  # Allow null for non-OAuth tokens
    )
    # FIXME: refresh_tokens not implemented
    # should be a relation to another token table
    # refresh_token = Column(
    #     Integer,
    #     ForeignKey('refresh_tokens.id', ondelete="CASCADE"),
    #     nullable=True,
    # )

    # the browser session id associated with a given token,
    # if issued during oauth to be stored in a cookie
    session_id = Column(Unicode(255), nullable=True)
    now = datetime.now(timezone.utc)
    created = Column(DateTime, default=datetime.now(timezone.utc))
    expires_at = Column(DateTime, default=None, nullable=True)
    last_activity = Column(DateTime)
    note = Column(Unicode(1023))
    scopes = Column(JSONList, default=[])

    def __repr__(self):
        kind = "user"
        name = self.user.name if self.user else None
        return "<{cls}('{pre}...', {kind}='{name}', client_id={client_id!r})>".format(
            cls=self.__class__.__name__,
            pre=self.prefix,
            kind=kind,
            name=name,
            client_id=self.client_id,
        )

    @classmethod
    def find(cls, db, token):
        """Find a token object by value. Returns None if not found."""
        prefix_match = cls.find_prefix(db, token)
        prefix_match = prefix_match.filter(cls.user_id.isnot(None))
        for orm_token in prefix_match:
            if orm_token.match(token):
                # Only purge if it's explicitly broken (has no client *and* shouldn't exist)
                # But we now support client-less tokens.
                return orm_token

    @classmethod
    def new(
        cls,
        token=None,
        *,
        user=None,
        scopes=None,
        note="",
        generated=True,
        session_id=None,
        expires_in=None,
        client_id=None,
        oauth_client=None,
        return_orm=False,
    ):
        """Generate a new API token for a user"""
        assert user
        db = inspect(user).session
        if token is None:
            token = new_token()
            generated = True
        else:
            cls.check_token(db, token)

        if oauth_client is None and client_id is not None:
            # Lazy-load only if a specific client_id was provided
            from .oauthclient import OAuthClient

            oauth_client = db.query(OAuthClient).filter_by(identifier=client_id).one_or_none()

        if oauth_client:
            client_id = oauth_client.identifier

        orm_token = cls(
            generated=generated,
            note=note or "",
            client_id=client_id,
            session_id=session_id,
            scopes=scopes if scopes else [],
        )
        db.add(orm_token)
        orm_token.token = token
        orm_token.user = user

        if expires_in is not None:
            orm_token.expires_at = cls.now + timedelta(seconds=expires_in)

        db.commit()
        return orm_token if return_orm else token



import argparse
import os

from grader_service.autograding.celery.app import CeleryApp


def main():
    parser = argparse.ArgumentParser(
        prog="Grader Worker", description="Starts celery worker for grader service."
    )
    parser.add_argument("-f", "--config", help="config file path", required=True)

    args = parser.parse_args()

    celery = CeleryApp.instance(config_file=os.path.abspath(args.config))
    app = celery.app

    worker = app.Worker(**celery.worker_kwargs)
    worker.start()


if __name__ == "__main__":
    main()



# Copyright (c) 2022, TU Wien
# All rights reserved.
# grader service orm
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from datetime import UTC, datetime
from enum import StrEnum

from sqlalchemy import Boolean, Column, DateTime, Enum, Float, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from grader_service.api.models import submission
from grader_service.orm.base import Base, DeleteState, Serializable


class AutoStatus(StrEnum):
    """Allowed `auto_status` values of a submission"""

    NOT_GRADED = "not_graded"
    PENDING = "pending"
    AUTOMATICALLY_GRADED = "automatically_graded"
    GRADING_FAILED = "grading_failed"


class ManualStatus(StrEnum):
    """Allowed `manual_status` values of a submission"""

    NOT_GRADED = "not_graded"
    MANUALLY_GRADED = "manually_graded"
    BEING_EDITED = "being_edited"


class FeedbackStatus(StrEnum):
    """Allowed `feedback_status` values of a submission"""

    NOT_GENERATED = "not_generated"
    GENERATING = "generating"
    GENERATED = "generated"
    GENERATION_FAILED = "generation_failed"
    FEEDBACK_OUTDATED = "feedback_outdated"


class Submission(Base, Serializable):
    __tablename__ = "submission"
    id = Column(Integer, primary_key=True, autoincrement=True)
    date = Column(DateTime, nullable=False)
    auto_status = Column(Enum(AutoStatus), default=AutoStatus.NOT_GRADED, nullable=False)
    manual_status = Column(Enum(ManualStatus), default=ManualStatus.NOT_GRADED, nullable=False)
    score = Column(Float, nullable=True)
    assignid = Column(Integer, ForeignKey("assignment.id", ondelete="CASCADE"))
    user_id = Column(Integer, ForeignKey("user.id", ondelete="CASCADE"), nullable=False)
    commit_hash = Column(String(length=40), nullable=False)
    feedback_status = Column(
        Enum(FeedbackStatus), default=FeedbackStatus.NOT_GENERATED, nullable=False
    )
    deleted = Column(Enum(DeleteState), nullable=False, unique=False, default=DeleteState.active)
    edited = Column(Boolean, nullable=False)
    updated_at = Column(
        DateTime, default=datetime.now(UTC), onupdate=datetime.now(UTC), nullable=False
    )
    grading_score = Column(Float, nullable=False)
    score_scaling = Column(Float, server_default="1.0", nullable=False)

    assignment = relationship("Assignment", back_populates="submissions")
    user = relationship("User", back_populates="submissions")
    logs = relationship(
        "SubmissionLogs",
        back_populates="submission",
        uselist=False,
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    properties = relationship(
        "SubmissionProperties",
        back_populates="submission",
        uselist=False,
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    @hybrid_property
    def user_display_name(self) -> str:
        return self.user.display_name

    @property
    def model(self) -> submission.Submission:
        model = submission.Submission(
            id=self.id,
            submitted_at=None if self.date is None else (self.date.isoformat("T", "milliseconds")),
            user_id=self.user_id,
            user_display_name=self.user_display_name,
            auto_status=self.auto_status,
            manual_status=self.manual_status,
            score_scaling=self.score_scaling,
            grading_score=self.grading_score,
            score=self.score,
            assignid=self.assignid,
            commit_hash=self.commit_hash,
            feedback_status=self.feedback_status,
            edited=self.edited,
        )
        return model

    def serialize_with_user(self) -> dict:
        """Serialize the submission with user information.

        Returns:
            dict: The serialized submission data including user information.
        """
        model = self.model.to_dict()
        model["user"] = self.user.serialize()
        return model



"""add user PK id

Revision ID: 9983ef1fda76
Revises: f1ae66d52ad9
Create Date: 2025-07-16 18:36:33.564133

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9983ef1fda76"
down_revision = "597857864aed"
branch_labels = None
depends_on = None


def upgrade():
    user_table = sa.table("user", sa.column("id"), sa.column("name"))
    dialect = op.get_bind().dialect.name

    if dialect == "sqlite":
        # sqlite has to recreate the tables on `batch_alter_table`, but dropping a table
        # would cause integrity errors, so we disable the foreign key constraint temporarily
        op.execute(sa.text("PRAGMA foreign_keys=OFF"))

    # 0. Drop FKs referencing user.name
    if dialect == "postgresql":
        for table, fk in [
            ("takepart", "takepart_username_fkey"),
            ("submission", "submission_username_fkey"),
            ("api_token", "api_token_username_fkey"),
            ("oauth_code", "oauth_code_username_fkey"),
        ]:
            with op.batch_alter_table(table) as batch_op:
                batch_op.drop_constraint(fk, type_="foreignkey")

    # 1. Switch PK from name -> id
    try:
        with op.batch_alter_table("user") as batch_op:
            batch_op.drop_constraint("user_pkey", type_="primary")
    except ValueError:
        pass  # sqlite does not name PK constraints

    with op.batch_alter_table("user") as batch_op:
        if dialect == "postgresql":
            batch_op.add_column(sa.Column("id", sa.Integer(), primary_key=True, autoincrement=True))
        elif dialect == "sqlite":
            # For sqlite, cannot add the PK constraint on column creation, because it causes
            # an IntegrityError (NOT NULL) in the temporary table.
            batch_op.add_column(sa.Column("id", sa.Integer(), nullable=False))

        batch_op.create_primary_key("pk_user_id", ["id"])
        batch_op.create_unique_constraint("unique_user_name", ["name"])

    def _add_user_id_col(table_name: str) -> None:
        op.add_column(table_name, sa.Column("user_id", sa.Integer(), nullable=True))
        table = sa.table(table_name, sa.column("username"), sa.column("user_id"))
        user_subq = (
            sa.select(user_table.c.id)
            .where(user_table.c.name == table.c.username)
            .scalar_subquery()
        )
        op.execute(table.update().values(user_id=user_subq))
        with op.batch_alter_table(table_name) as batch_op:
            batch_op.alter_column("user_id", nullable=False)

    # 2. takepart
    _add_user_id_col("takepart")
    with op.batch_alter_table("takepart") as batch_op:
        batch_op.drop_column("username")
        batch_op.create_foreign_key("fk_takepart_user_id", "user", ["user_id"], ["id"])
        batch_op.create_primary_key("pk_takepart", ["user_id", "lectid"])

    # 3. submission
    _add_user_id_col("submission")
    with op.batch_alter_table("submission") as batch_op:
        batch_op.create_foreign_key("fk_submission_user_id", "user", ["user_id"], ["id"])
        batch_op.drop_column("username")

    # 4. api_token
    _add_user_id_col("api_token")
    with op.batch_alter_table("api_token") as batch_op:
        batch_op.create_foreign_key("fk_api_token_user_id", "user", ["user_id"], ["id"])
        batch_op.drop_column("username")

    # 5. oauth_code
    _add_user_id_col("oauth_code")
    with op.batch_alter_table("oauth_code") as batch_op:
        batch_op.create_foreign_key("fk_oauth_code_user_id", "user", ["user_id"], ["id"])
        batch_op.drop_column("username")


def downgrade():
    dialect = op.get_bind().dialect.name
    if dialect == "sqlite":
        # sqlite has to recreate the tables on `batch_alter_table`, but dropping a table
        # would cause integrity errors, so we disable the foreign key constraint temporarily
        op.execute(sa.text("PRAGMA foreign_keys=OFF"))

    user_table = sa.table("user", sa.column("id"), sa.column("name"))

    def _add_username_col(table_name: str) -> None:
        """Adds a username column, fills it based on user_id, and enforces NOT NULL."""
        op.add_column(table_name, sa.Column("username", sa.String(length=255), nullable=True))

        table = sa.table(table_name, sa.column("username"), sa.column("user_id"))
        user_subq = (
            sa.select(user_table.c.name).where(user_table.c.id == table.c.user_id).scalar_subquery()
        )
        op.execute(table.update().values(username=user_subq))

    # 5. oauth_code
    _add_username_col("oauth_code")
    with op.batch_alter_table("oauth_code") as batch_op:
        batch_op.drop_constraint("fk_oauth_code_user_id", type_="foreignkey")
        batch_op.drop_column("user_id")

    # 4. api_token
    _add_username_col("api_token")
    with op.batch_alter_table("api_token") as batch_op:
        batch_op.drop_constraint("fk_api_token_user_id", type_="foreignkey")
        batch_op.drop_column("user_id")

    # 3. submission
    _add_username_col("submission")
    with op.batch_alter_table("submission") as batch_op:
        batch_op.drop_constraint("fk_submission_user_id", type_="foreignkey")
        batch_op.drop_column("user_id")

    # 2. takepart
    _add_username_col("takepart")
    with op.batch_alter_table("takepart") as batch_op:
        batch_op.drop_constraint("fk_takepart_user_id", type_="foreignkey")
        batch_op.drop_constraint("pk_takepart", type_="primary")
        batch_op.create_primary_key("pk_takepart", ["username", "lectid"])
        batch_op.drop_column("user_id")
        # Note: Only "takepart" has the constraint `nullable=False` on the "username" column (!).
        batch_op.alter_column("username", nullable=False)

    # 1. Switch PK back from id -> name
    with op.batch_alter_table("user") as batch_op:
        batch_op.drop_constraint("pk_user_id", type_="primary")
        batch_op.drop_constraint("unique_user_name", type_="unique")
        batch_op.drop_column("id")
        batch_op.create_primary_key("user_pkey", ["name"])  # restore name as PK

    # 0. Create FKs referencing user.name
    for table, fk in [
        ("takepart", "takepart_username_fkey"),
        ("submission", "submission_username_fkey"),
        ("api_token", "api_token_username_fkey"),
        ("oauth_code", "oauth_code_username_fkey"),
    ]:
        with op.batch_alter_table(table) as batch_op:
            batch_op.create_foreign_key(fk, "user", ["username"], ["name"])



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import fnmatch
import json
import os
import shutil
import subprocess
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from sqlalchemy.orm import Session
from traitlets.config import Config
from traitlets.config.configurable import LoggingConfigurable
from traitlets.traitlets import Int, TraitError, Type, Unicode, validate

from grader_service.autograding.git_manager import GitSubmissionManager
from grader_service.autograding.utils import collect_logs, executable_validator, rmtree
from grader_service.convert.converters.autograde import Autograde
from grader_service.convert.gradebook.models import GradeBookModel
from grader_service.orm.assignment import Assignment
from grader_service.orm.submission import AutoStatus, ManualStatus, Submission
from grader_service.orm.submission_logs import SubmissionLogs
from grader_service.orm.submission_properties import SubmissionProperties


class LocalAutogradeExecutor(LoggingConfigurable):
    """
    Runs an autograde job on the local machine
    with the current Python environment.
    Sets up the necessary directories
    and the gradebook JSON file used by :mod:`grader_service.convert`.
    """

    relative_input_path = Unicode("convert_in", allow_none=True).tag(config=True)
    relative_output_path = Unicode("convert_out", allow_none=True).tag(config=True)
    git_manager_class = Type(GitSubmissionManager, allow_none=False).tag(config=True)

    cell_timeout = Int(
        allow_none=False,
        help="Returns the cell timeout in seconds, either user-defined, from configuration or default values.",
    ).tag(config=True)

    default_cell_timeout = Int(300, help="Default cell timeout in seconds, defaults to 300").tag(
        config=True
    )

    min_cell_timeout = Int(10, help="Min cell timeout in seconds, defaults to 10.").tag(config=True)

    max_cell_timeout = Int(
        86400, help="Max cell timeout in seconds, defaults to 86400 (24 hours)"
    ).tag(config=True)

    def __init__(
        self, grader_service_dir: str, submission: Submission, close_session: bool = True, **kwargs
    ):
        """
        Creates the executor in the input
        and output directories that are specified
        by :attr:`base_input_path` and :attr:`base_output_path`.
        The grader service directory is used for accessing
        the git repositories to push the grading results.
        The database session is retrieved from the submission object.
        The associated session of the submission has to be available
        and must not be closed beforehand.

        :param grader_service_dir: The base directory of the whole
        grader service specified in the configuration.
        :type grader_service_dir: str
        :param submission: The submission object
        which should be graded by the executor.
        :type submission: Submission
        :param close_session: Whether to close the db session after grading.
        :type close_session: bool
        """
        super().__init__(**kwargs)
        self.grader_service_dir = grader_service_dir
        self.submission = submission
        self.assignment: Assignment = submission.assignment
        self.session: Session = Session.object_session(self.submission)
        # close session after grading (might need session later)
        self.close_session = close_session

        self.grading_logs: Optional[str] = None
        # Git manager performs the git operations when creating a new repo for the grading results
        self.git_manager = self.git_manager_class(grader_service_dir, self.submission)

        self.cell_timeout = self._determine_cell_timeout()

    def start(self):
        """
        Starts the autograding job.
        This is the only method that is exposed to the client.
        """
        self.log.info(
            "Starting autograding job for submission %s in %s",
            self.submission.id,
            self.__class__.__name__,
        )
        try:
            self._clean_up_input_and_output_dirs()
            self.git_manager.pull_submission(self.input_path)

            autograding_start = datetime.now()
            self._write_gradebook(self._put_grades_in_assignment_properties())
            self._run()
            autograding_finished = datetime.now()

            files_to_commit = self._get_whitelisted_files()
            self.git_manager.push_results(files_to_commit, self.output_path)
            self._set_properties()
            self._set_db_state()
        except Exception as e:
            self.log.error(
                "Failed autograding job for submission %s in %s",
                self.submission.id,
                self.__class__.__name__,
                exc_info=True,
            )
            if isinstance(e, subprocess.CalledProcessError):
                err_msg = e.stderr
            else:
                err_msg = str(e)
            self.grading_logs = (self.grading_logs or "") + err_msg
            self._set_db_state(success=False)
        else:
            ts = round((autograding_finished - autograding_start).total_seconds())
            self.log.info(
                "Successfully completed autograding job for submission %s in %s; took %s min %s s",
                self.submission.id,
                self.__class__.__name__,
                ts // 60,
                ts % 60,
            )
        finally:
            self._update_submission_logs()
            self._cleanup()

    @property
    def input_path(self):
        return os.path.join(
            self.grader_service_dir, self.relative_input_path, f"submission_{self.submission.id}"
        )

    @property
    def output_path(self):
        return os.path.join(
            self.grader_service_dir, self.relative_output_path, f"submission_{self.submission.id}"
        )

    def _clean_up_input_and_output_dirs(self):
        """Prepare clean input and output dirs before the autograde process"""
        if os.path.exists(self.input_path):
            rmtree(self.input_path)
        os.makedirs(self.input_path)
        if os.path.exists(self.output_path):
            rmtree(self.output_path)
        os.makedirs(self.output_path)

    def _write_gradebook(self, gradebook_str: str) -> None:
        """
        Writes the gradebook to the output directory where it will be used by
        :mod:`grader_service.convert` to load the data.
        The name of the written file is gradebook.json.
        :param gradebook_str: The content of the gradebook.
        :return: None
        """
        path = os.path.join(self.output_path, "gradebook.json")
        self.log.info(f"Writing gradebook to {path}")
        with open(path, "w") as f:
            f.write(gradebook_str)

    def _run(self):
        """
        Runs the autograding in the current interpreter
        and captures the output.
        """
        autograder = Autograde(
            self.input_path,
            self.output_path,
            "*.ipynb",
            assignment_settings=self.assignment.settings,
            config=self._get_autograde_config(),
        )

        # Add a handler to the autograder's logger so that we can capture its logs and write them
        # to self.grading_logs:
        with collect_logs(autograder.log) as log_stream:
            autograder.start()
            self.grading_logs = log_stream.getvalue()

    def _put_grades_in_assignment_properties(self) -> str:
        """
        Checks if assignment was already graded and returns updated properties.
        :return: str
        """
        if self.submission.manual_status == ManualStatus.NOT_GRADED:
            return self.assignment.properties

        assignment_properties = json.loads(self.assignment.properties)
        submission_properties = json.loads(self.submission.properties.properties)
        notebooks = set.intersection(
            set(assignment_properties["notebooks"].keys()),
            set(submission_properties["notebooks"].keys()),
        )
        for notebook in notebooks:
            # Set grades
            assignment_properties["notebooks"][notebook]["grades_dict"] = submission_properties[
                "notebooks"
            ][notebook]["grades_dict"]
            # Set comments
            assignment_properties["notebooks"][notebook]["comments_dict"] = submission_properties[
                "notebooks"
            ][notebook]["comments_dict"]

        self.log.info("Added grades dict to properties")
        properties_str = json.dumps(assignment_properties)
        return properties_str

    def _get_autograde_config(self) -> Config:
        """Returns the autograde config, with the timeout set for ExecutePreprocessor."""
        c = Config()
        c.ExecutePreprocessor.timeout = self.cell_timeout
        return c

    def _get_whitelist_patterns(self) -> set[str]:
        """Return the glob patterns which are used for whitelisting the generated autograded files."""
        return self.assignment.get_whitelist_patterns()

    def _get_whitelisted_files(self) -> List[str]:
        """
        Prepares a list of shell-escaped filenames matching the whitelist patterns of the assignment.

        The list can be directly passed to the `git commit` command.

        :return: list of shell-escaped filenames matching the whitelist patterns of the assignment
        """
        file_patterns = self._get_whitelist_patterns()
        if not file_patterns:
            # No filtering needed
            return ["."]

        files_to_commit = []

        # get all files in the directory
        for root, dirs, files in os.walk(self.output_path):
            # Exclude .git directory - it contains subdirectories which we don't need to check
            if ".git" in root:
                continue
            rel_root = os.path.relpath(root, self.output_path)
            for file in files:
                file_path = os.path.join(rel_root, file) if rel_root != "." else file
                if any(fnmatch.fnmatch(file_path, pattern) for pattern in file_patterns):
                    files_to_commit.append(file_path)

        return files_to_commit

    def _set_properties(self) -> None:
        """
        Loads the contents of the gradebook.json file
        and sets them as the submission properties.
        Also calculates the score of the submission
        after autograding based on the updated properties.

        :return: None
        """
        with open(os.path.join(self.output_path, "gradebook.json"), "r") as f:
            gradebook_str = f.read()

        properties = SubmissionProperties(properties=gradebook_str, sub_id=self.submission.id)

        self.session.merge(properties)

        gradebook_dict = json.loads(gradebook_str)
        book = GradeBookModel.from_dict(gradebook_dict)
        score = 0
        for id, n in book.notebooks.items():
            score += n.score
        self.submission.grading_score = score
        self.submission.score = self.submission.score_scaling * score
        self.session.commit()

    def _set_db_state(self, success=True) -> None:
        """
        Sets the submission autograding status based on the success parameter
        and sets the logs from autograding.

        :param success: Whether the grading process was a success or failure.
        :return: None
        """
        if success:
            self.submission.auto_status = AutoStatus.AUTOMATICALLY_GRADED
        else:
            self.submission.auto_status = AutoStatus.GRADING_FAILED
        self.session.commit()

    def _update_submission_logs(self):
        if self.grading_logs is not None:
            # Remove null characters to avoid database storage/string processing/display/etc. issues
            self.grading_logs = self.grading_logs.replace("\x00", "")
        logs = SubmissionLogs(logs=self.grading_logs, sub_id=self.submission.id)
        self.session.merge(logs)
        self.session.commit()

    def _cleanup(self) -> None:
        """
        Removes all files from the input and output directories
        and closes the session if specified by self.close_session.

        Note: This also removes the gradebook.json file, which is in the `self.output_path` dir.
        :return: None
        """
        try:
            shutil.rmtree(self.input_path)
            shutil.rmtree(self.output_path)
        except FileNotFoundError:
            pass
        if self.close_session:
            self.session.close()

    def _determine_cell_timeout(self):
        cell_timeout = self.default_cell_timeout
        # check if the cell timeout was set by user
        if self.assignment.settings.cell_timeout is not None:
            custom_cell_timeout = self.assignment.settings.cell_timeout
            self.log.info(
                f"Found custom cell timeout in assignment settings: {custom_cell_timeout} seconds."
            )
            cell_timeout = min(
                self.max_cell_timeout, max(custom_cell_timeout, self.min_cell_timeout)
            )
            self.log.info(f"Setting custom cell timeout to {cell_timeout}.")

        return cell_timeout

    @validate("min_cell_timeout", "default_cell_timeout", "max_cell_timeout")
    def _validate_cell_timeouts(self, proposal):
        trait_name = proposal["trait"].name
        value = proposal["value"]

        # Get current or proposed values
        min_t = value if trait_name == "min_cell_timeout" else self.min_cell_timeout
        default_t = value if trait_name == "default_cell_timeout" else self.default_cell_timeout
        max_t = value if trait_name == "max_cell_timeout" else self.max_cell_timeout

        # Validate the constraint
        if not (0 < min_t < default_t < max_t):
            raise TraitError(
                f"Invalid {trait_name} value ({value}). "
                f"Timeout values must satisfy: 0 < min_cell_timeout < default_cell_timeout < max_cell_timeout. "
                f"Got min={min_t}, default={default_t}, max={max_t}."
            )

        return value

    @validate("relative_input_path", "relative_output_path")
    def _validate_service_dir(self, proposal):
        path: str = proposal["value"]
        if not os.path.exists(self.grader_service_dir + "/" + path):
            self.log.info(f"Path {path} not found, creating new directories.")
            Path(path).mkdir(parents=True, exist_ok=True, mode=0o700)
        if not os.path.isdir(self.grader_service_dir + "/" + path):
            raise TraitError("The path has to be an existing directory")
        return path

    @validate("convert_executable")
    def _validate_executable(self, proposal):
        return executable_validator(proposal)


class LocalAutogradeProcessExecutor(LocalAutogradeExecutor):
    """Runs an autograde job on the local machine
    with the default Python environment in a separate process.
    Sets up the necessary directories
    and the gradebook JSON file used by :mod:`grader_service.convert`.
    """

    convert_executable = Unicode("grader-convert", allow_none=False).tag(config=True)

    def _run(self):
        """
        Runs the autograding in a separate python interpreter
        as a sub-process and captures the output.
        """
        command = [
            self.convert_executable,
            "autograde",
            "-i",
            self.input_path,
            "-o",
            self.output_path,
            "-p",
            "*.ipynb",
            f"--ExecutePreprocessor.timeout={self.cell_timeout}",
        ]
        self.log.info(f"Running {command}")
        process = subprocess.run(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=None, text=True
        )
        self.grading_logs = process.stderr
        if process.returncode == 0:
            self.log.info(self.grading_logs)
            self.log.info("Process has successfully completed execution!")
        else:
            self.log.error(self.grading_logs)
            raise RuntimeError("Process has failed execution!")
